import numpy as np

from landlab.components.erosion_deposition.generalized_erosion_deposition import (
    DEFAULT_MINIMUM_TIME_STEP,
    _GeneralizedErosionDeposition,
)
from landlab.utils.return_array import return_array_at_node

from .cfuncs import calculate_qs_in

ROOT2 = np.sqrt(2.0)  # syntactic sugar for precalculated square root of 2
TIME_STEP_FACTOR = 0.5  # factor used in simple subdivision solver


class ErosionDeposition(_GeneralizedErosionDeposition):
    r"""
    Erosion-Deposition model in the style of Davy and Lague (2009). It uses a
    mass balance approach across the total sediment mass both in the bed and
    in transport coupled with explicit representation of the sediment
    transport lengthscale (the "xi-q" model) to derive a range of erosional
    and depositional responses in river channels.

    This implementation is close to the Davy & Lague scheme, with a few
    deviations:

        - A fraction of the eroded sediment is permitted to enter the wash load,
          and lost to the mass balance (`F_f`).

        - Here an incision threshold :math:`\omega` is permitted, where it was not by Davy &
          Lague. It is implemented with an exponentially smoothed form to prevent
          discontinuities in the parameter space. See the
          :py:class:`~landlab.components.StreamPowerSmoothThresholdEroder`
          for more documentation.

        - This component uses an "effective" settling velocity, v_s, as one of its
          inputs. This parameter is simply equal to Davy & Lague's `d_star * V`
          dimensionless number.

    Erosion of the bed follows a stream power formulation, i.e.,

    .. math:

        E = K * q ** m_{sp} * S ** {n_sp} - \omega

    Note that the transition between transport-limited and detachment-limited
    behavior is controlled by the dimensionless ratio (v_s/r) where r is the
    runoff ratio (Q=Ar). r can be changed in the flow accumulation component
    but is not changed within ErosionDeposition. Because the runoff ratio r
    is not changed within the ErosionDeposition component,  v_s becomes the
    parameter that fundamentally controls response style. Very small v_s will
    lead to a detachment-limited response style, very large v_s will lead to a
    transport-limited response style. v_s == 1 means equal contributions from
    transport and erosion, and a hybrid response as described by Davy & Lague.

    Unlike other some other fluvial erosion componets in Landlab, in this
    component (and :py:class:`~landlab.components.SPACE`) no erosion occurs
    in depressions or in areas with adverse slopes. There is no ability to
    pass a keyword argument ``erode_flooded_nodes``.

    If a depressions are handled (as indicated by the presence of the field
    "flood_status_code" at nodes), then deposition occurs throughout the
    depression and sediment is passed out of the depression. Where pits are
    encountered, then all sediment is deposited at that node only.

    A note about sediment porosity: Prior to Landlab v2.0 this component took a
    porositiy keyworkd argument ``phi``. For an explaination of why it no
    longer does (including a mathematical derivation), see
    `Pull Request 1186 <https://github.com/landlab/landlab/pull/1186>`_.
    If ``phi`` is passed to this component a value error will be raised.

    Component written by C. Shobe, K. Barnhart, and G. Tucker.

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Barnhart, K., Glade, R., Shobe, C., Tucker, G. (2019). Terrainbento 1.0: a
    Python package for multi-model analysis in long-term drainage basin
    evolution. Geoscientific Model Development  12(4), 1267--1297.
    https://dx.doi.org/10.5194/gmd-12-1267-2019

    **Additional References**

    Davy, P., Lague, D. (2009). Fluvial erosion/transport equation of landscape
    evolution models revisited Journal of Geophysical Research  114(F3),
    F03007. https://dx.doi.org/10.1029/2008jf001146

    """

    _name = "ErosionDeposition"

    _unit_agnostic = True

    _cite_as = """
    @article{barnhart2019terrain,
      author = {Barnhart, Katherine R and Glade, Rachel C and Shobe, Charles M and Tucker, Gregory E},
      title = {{Terrainbento 1.0: a Python package for multi-model analysis in long-term drainage basin evolution}},
      doi = {10.5194/gmd-12-1267-2019},
      pages = {1267---1297},
      number = {4},
      volume = {12},
      journal = {Geoscientific Model Development},
      year = {2019},
    }
    """

    _info = {
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "sediment__flux": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m3/s",
            "mapping": "node",
            "doc": "Sediment flux (volume per unit time of sediment entering each node)",
        },
        "surface_water__discharge": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m**2/s",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
    }

    def __init__(
        self,
        grid,
        K=0.002,
        v_s=1.0,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit=0.0,
        F_f=0.0,
        discharge_field="surface_water__discharge",
        solver="basic",
        dt_min=DEFAULT_MINIMUM_TIME_STEP,
        **kwds
    ):
        """Initialize the ErosionDeposition model.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        K : float, field name, or array
            Erodibility for substrate (units vary).
        v_s : float
            Effective settling velocity for chosen grain size metric [L/T].
        m_sp : float
            Discharge exponent (units vary)
        n_sp : float
            Slope exponent (units vary)
        sp_crit : float, field name, or array
            Critical stream power to erode substrate [E/(TL^2)]
        F_f : float
            Fraction of eroded material that turns into "fines" that do not
            contribute to (coarse) sediment load. Defaults to zero.
        discharge_field : float, field name, or array
            Discharge [L^2/T]. The default is to use the grid field
            'surface_water__discharge', which is simply drainage area
            multiplied by the default rainfall rate (1 m/yr). To use custom
            spatially/temporally varying rainfall, use 'water__unit_flux_in'
            to specify water input to the FlowAccumulator.
        solver : string
            Solver to use. Options at present include:
                (1) 'basic' (default): explicit forward-time extrapolation.
                    Simple but will become unstable if time step is too large.
                (2) 'adaptive': adaptive time-step solver that estimates a
                    stable step size based on the shortest time to "flattening"
                    among all upstream-downstream node pairs.

        Examples
        ---------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> from landlab.components import DepressionFinderAndRouter
        >>> from landlab.components import ErosionDeposition
        >>> from landlab.components import FastscapeEroder
        >>> np.random.seed(seed = 5000)

        Define grid and initial topography:
            -5x5 grid with baselevel in the lower left corner
            -all other boundary nodes closed
            -Initial topography is plane tilted up to the upper right + noise

        >>> nr = 5
        >>> nc = 5
        >>> dx = 10
        >>> mg = RasterModelGrid((nr, nc), xy_spacing=10.0)
        >>> _ = mg.add_zeros('node', 'topographic__elevation')
        >>> mg['node']['topographic__elevation'] += (mg.node_y/10 +
        ...        mg.node_x/10 + np.random.rand(len(mg.node_y)) / 10)
        >>> mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
        ...                                               left_is_closed=True,
        ...                                               right_is_closed=True,
        ...                                               top_is_closed=True)
        >>> mg.set_watershed_boundary_condition_outlet_id(0,\
                mg['node']['topographic__elevation'], -9999.)
        >>> fsc_dt = 100.
        >>> ed_dt = 1.

        Check initial topography

        >>> mg.at_node['topographic__elevation'] # doctest: +NORMALIZE_WHITESPACE
        array([ 0.02290479,  1.03606698,  2.0727653 ,  3.01126678,  4.06077707,
            1.08157495,  2.09812694,  3.00637448,  4.07999597,  5.00969486,
            2.04008677,  3.06621577,  4.09655859,  5.04809001,  6.02641123,
            3.05874171,  4.00585786,  5.0595697 ,  6.04425233,  7.05334077,
            4.05922478,  5.0409473 ,  6.07035008,  7.0038935 ,  8.01034357])

        Instantiate Fastscape eroder, flow router, and depression finder

        >>> fr = FlowAccumulator(mg, flow_director='D8')
        >>> df = DepressionFinderAndRouter(mg)
        >>> fsc = FastscapeEroder(
        ...     mg,
        ...     K_sp=.001,
        ...     m_sp=.5,
        ...     n_sp=1)

        Burn in an initial drainage network using the Fastscape eroder:

        >>> for x in range(100):
        ...     fr.run_one_step()
        ...     df.map_depressions()
        ...     flooded = np.where(df.flood_status==3)[0]
        ...     fsc.run_one_step(dt = fsc_dt)
        ...     mg.at_node['topographic__elevation'][0] -= 0.001 #uplift

        Instantiate the E/D component:

        >>> ed = ErosionDeposition(
        ...     mg,
        ...     K=0.00001,
        ...     v_s=0.001,
        ...     m_sp=0.5,
        ...     n_sp = 1.0,
        ...     sp_crit=0)

        Now run the E/D component for 2000 short timesteps:

        >>> for x in range(2000): #E/D component loop
        ...     fr.run_one_step()
        ...     df.map_depressions()
        ...     ed.run_one_step(dt = ed_dt)
        ...     mg.at_node['topographic__elevation'][0] -= 2e-4 * ed_dt

        Now we test to see if topography is right:

        >>> np.around(mg.at_node['topographic__elevation'], decimals=3) # doctest: +NORMALIZE_WHITESPACE
        array([-0.477,  1.036,  2.073,  3.011,  4.061,  1.082, -0.08 , -0.065,
           -0.054,  5.01 ,  2.04 , -0.065, -0.065, -0.053,  6.026,  3.059,
           -0.054, -0.053, -0.035,  7.053,  4.059,  5.041,  6.07 ,  7.004,
            8.01 ])
        """
        if grid.at_node["flow__receiver_node"].size != grid.size("node"):
            msg = (
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that ErosionDeposition is compatible with "
                "route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )
            raise NotImplementedError(msg)

        if "phi" in kwds:
            msg = "As of Landlab v2 ErosionDeposition no longer takes the keyword argument phi. The sediment flux is considered to represent bulk deposit volume rather than mineral volume, and therefore porosity does not impact the dynamics. The following pull request explains the math behind this: https://github.com/landlab/landlab/pull/1186."
            raise ValueError(msg)
        elif len(kwds) > 0:
            kwdstr = " ".join(list(kwds.keys()))
            raise ValueError(
                "Extra kwds passed to ErosionDeposition:{kwds}".format(kwds=kwdstr)
            )
        super().__init__(
            grid,
            m_sp=m_sp,
            n_sp=n_sp,
            F_f=F_f,
            v_s=v_s,
            dt_min=dt_min,
            discharge_field=discharge_field,
        )

        # E/D specific inits.

        # K's and critical values can be floats, grid fields, or arrays
        # use setter for K defined below
        self.K = K
        self._sp_crit = return_array_at_node(grid, sp_crit)

        # Handle option for solver
        if solver == "basic":
            self.run_one_step = self.run_one_step_basic
        elif solver == "adaptive":
            self.run_one_step = self.run_with_adaptive_time_step_solver
            self._time_to_flat = np.zeros(grid.number_of_nodes)
        else:
            raise ValueError(
                "Parameter 'solver' must be one of: " + "'basic', 'adaptive'"
            )

    @property
    def K(self):
        """Erodibility (units depend on m_sp)."""
        return self._K

    @K.setter
    def K(self, new_val):
        self._K = return_array_at_node(self._grid, new_val)

    def _calc_erosion_rates(self):
        """Calculate erosion rates."""
        omega = self._K * self._Q_to_the_m * np.power(self._slope, self._n_sp)
        omega_over_sp_crit = np.divide(
            omega, self._sp_crit, out=np.zeros_like(omega), where=self._sp_crit != 0
        )

        self._erosion_term = omega - self._sp_crit * (1.0 - np.exp(-omega_over_sp_crit))

    def _calc_qs_in_and_depo_rate(self):
        self._calc_hydrology()
        self._calc_erosion_rates()

        is_flooded_core_node = self._get_flooded_core_nodes()
        self._erosion_term[is_flooded_core_node] = 0.0

        self._qs_in[:] = 0.0
        self._depo_rate[:] = 0.0

        # iterate top to bottom through the stack, calculate qs
        # cythonized version of calculating qs_in
        calculate_qs_in(
            np.flipud(self._stack),
            self._flow_receivers,
            self._cell_area_at_node,
            self._q,
            self._qs,
            self._qs_in,
            self._erosion_term,
            self._v_s,
            self._F_f,
        )

        self._depo_rate[self._q > 0] = self._qs[self._q > 0] * (
            self._v_s / self._q[self._q > 0]
        )
        if not self._depressions_are_handled():  # all sed dropped here
            self._depo_rate[is_flooded_core_node] = (
                self._qs_in[is_flooded_core_node]
                / self._cell_area_at_node[is_flooded_core_node]
            )

    def run_one_step_basic(self, dt=1.0):
        """Calculate change in rock and alluvium thickness for a time period
        'dt'.

        Parameters
        ----------
        dt : float
            Model timestep [T]
        """
        self._calc_qs_in_and_depo_rate()

        # topo elev is old elev + deposition - erosion
        cores = self._grid.core_nodes
        dzdt = self._depo_rate - self._erosion_term
        self._topographic__elevation[cores] += dzdt[cores] * dt

    def run_with_adaptive_time_step_solver(self, dt=1.0):
        """CHILD-like solver that adjusts time steps to prevent slope
        flattening.


        Parameters
        ----------
        dt : float
            Model timestep [T]
        """

        # Initialize remaining_time, which records how much of the global time
        # step we have yet to use up.
        remaining_time = dt

        z = self._grid.at_node["topographic__elevation"]
        r = self._flow_receivers
        dzdt = np.zeros(len(z))
        cores = self._grid.core_nodes

        first_iteration = True

        is_flooded_core_node = self._get_flooded_core_nodes()

        # Outer WHILE loop: keep going until time is used up
        while remaining_time > 0.0:

            # Update all the flow-link slopes.
            #
            # For the first iteration, we assume this has already been done
            # outside the component (e.g., by flow router), but we need to do
            # it ourselves on subsequent iterations.
            if not first_iteration:
                # update the link slopes
                self._update_flow_link_slopes()
                # update where nodes are flooded. This shouldn't happen bc
                # of the dynamic timestepper, but just in case, we update here.
                is_flooded_core_node[self._slope < 0] = True
            else:
                first_iteration = False

            self._calc_qs_in_and_depo_rate()

            # Rate of change of elevation at core nodes:
            dzdt[cores] = self._depo_rate[cores] - self._erosion_term[cores]

            # Difference in elevation between each upstream-downstream pair
            zdif = z - z[r]

            # Rate of change of the *difference* in elevation between each
            # upstream-downstream pair.
            rocdif = dzdt - dzdt[r]

            # (Re)-initialize the array that will contain "time to (almost)
            # flat" for each node (relative to its downstream neighbor).
            self._time_to_flat[:] = remaining_time

            # Find locations where the upstream and downstream node elevations
            # are converging (e.g., the upstream one is eroding faster than its
            # downstream neighbor)
            converging = np.nonzero(rocdif < 0.0)[0]

            # Find the time to (almost) flat by dividing difference by rate of
            # change of difference, and then multiplying by a "safety factor"
            self._time_to_flat[converging] = -(
                TIME_STEP_FACTOR * zdif[converging] / rocdif[converging]
            )

            # Mask out pairs where the source at the same or lower elevation
            # as its downstream neighbor (e.g., because it's a pit or a lake).
            # Here, masking out means simply assigning the remaining time in
            # the global time step.
            self._time_to_flat[np.nonzero(zdif <= 0.0)[0]] = remaining_time
            self._time_to_flat[is_flooded_core_node] = remaining_time

            # From this, find the maximum stable time step. If it is smaller
            # than our tolerance, report and quit.
            dt_max = max(np.amin(self._time_to_flat), self._dt_min)

            # Finally, apply dzdt to all nodes for a (sub)step of duration
            # dt_max
            z[cores] += dzdt[cores] * dt_max

            # Update remaining time and continue the loop
            remaining_time -= dt_max



# -*- coding: utf-8 -*-
"""
Created on Wed May 22 13:50:43 2019

@author: abby


Doc tests and unit tests for lateral erosion.
"""

import numpy as np
from numpy import testing

from landlab import RasterModelGrid
from landlab.components import FlowAccumulator, LateralEroder


def test_lateral_erosion_and_node():
    """
    Test that sets up a simple, pre-defined drainage network and compares
    the lateral node that is eroded, the volume of lateral eorsion, and the elevation
    of the landscape after one timestep to known values.
    """
    nr = 5
    nc = 5
    dx = 1
    mg = RasterModelGrid((nr, nc), xy_spacing=dx)
    for edge in (
        mg.nodes_at_top_edge,
        mg.nodes_at_bottom_edge,
        mg.nodes_at_left_edge,
        mg.nodes_at_right_edge,
    ):
        mg.status_at_node[edge] = mg.BC_NODE_IS_CLOSED
    for edge in mg.nodes_at_bottom_edge:
        mg.status_at_node[edge] = mg.BC_NODE_IS_FIXED_VALUE

    z = mg.add_zeros("topographic__elevation", at="node")
    loading_vector = np.linspace(1, 4, num=nr)
    ramp = np.repeat(loading_vector, nc)
    # the tweaks to elevation below make lateral node at node 7
    z += ramp
    z[11] -= 0.9
    z[12] -= 0.4
    z[8] -= 0.001
    fa = FlowAccumulator(
        mg,
        surface="topographic__elevation",
        flow_director="FlowDirectorD8",
        runoff_rate=None,
        depression_finder=None,
    )
    latero = LateralEroder(mg, latero_mech="UC", Kv=0.1, Kl_ratio=1.5)
    fa.accumulate_flow()
    (mg, dzlat) = latero.run_one_step(dt=1.0)

    vlname = mg["node"]["volume__lateral_erosion"]
    # predicted volume of lateral eorsion
    pred_vollat = 0.00045158164
    # predicted elevation after 1 time step
    pred_zafter = np.array(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.75,
            1.675,
            1.675,
            1.6731154,
            1.75,
            2.5,
            1.66418779,
            2.06181623,
            2.4249,
            2.5,
            3.25,
            3.085,
            3.13332738,
            3.16868272,
            3.25,
            4.0,
            4.0,
            4.0,
            4.0,
            4.0,
        ]
    )

    testing.assert_array_almost_equal(
        mg.at_node["topographic__elevation"],
        pred_zafter,
        decimal=8,
        err_msg="LatEro basic erosion test failed",
        verbose=True,
    )
    testing.assert_array_almost_equal(
        vlname[7],
        pred_vollat,
        decimal=8,
        err_msg="LatEro volume lateral erosion failed",
        verbose=True,
    )


def test_matches_detlim_solution():
    """
    Test that model matches the detachment-limited analytical solution
    for slope/area relationship at steady state: S=(U/K_br)^(1/n)*A^(-m/n).
    """

    U = 0.005
    dt = 10
    nr = 5
    nc = 5
    dx = 10
    Kbr = 0.001
    # instantiate grid
    mg = RasterModelGrid((nr, nc), xy_spacing=dx)
    for edge in (
        mg.nodes_at_top_edge,
        mg.nodes_at_bottom_edge,
        mg.nodes_at_left_edge,
        mg.nodes_at_right_edge,
    ):
        mg.status_at_node[edge] = mg.BC_NODE_IS_CLOSED
    for edge in mg.nodes_at_bottom_edge:
        mg.status_at_node[edge] = mg.BC_NODE_IS_FIXED_VALUE

    z = mg.add_zeros("topographic__elevation", at="node")
    ir2 = np.random.uniform(low=0.0, high=0.5, size=(z.size))
    loading_vector = np.linspace(1, 4, num=nr)
    ramp = np.repeat(loading_vector, nc)
    ramp += ir2
    z[:] += ramp
    n_sp = 1.0
    m_sp = 0.5

    fa = FlowAccumulator(
        mg,
        surface="topographic__elevation",
        flow_director="FlowDirectorD8",
        runoff_rate=None,
        depression_finder=None,
    )  # "DepressionFinderAndRouter", router="D8")

    # set alph = 0.0 to make run detachment limited and set Kl_ratio = 1.0 for
    # equal vertical and lateral bedrock erosion coefficients
    latero = LateralEroder(
        mg, latero_mech="TB", Kv=Kbr, solver="basic", alph=0.0, Kl_ratio=1.0
    )

    for i in range(2000):
        fa.run_one_step()  # flow accumulator
        # erode the landscape with lateral erosion
        (mg, dzlat) = latero.run_one_step(dt)
        mg.at_node["topographic__elevation"][mg.core_nodes] += U * dt

    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(U / Kbr, 1.0 / n_sp) * np.power(
        mg.at_node["drainage_area"][mg.core_nodes], -m_sp / n_sp
    )

    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=4,
        err_msg="LatEro detachment-limited steady state test failed",
        verbose=True,
    )


def test_ss_sed_flux():
    """
    Test that sediment flux of lateral erosion model matches steady state
    analytical solution, Qs = U * A
    """
    U = 0.0005  # m/yr
    dt = 100  # years
    nr = 5
    nc = 5
    dx = 10
    mg = RasterModelGrid((nr, nc), xy_spacing=dx)

    for edge in (
        mg.nodes_at_top_edge,
        mg.nodes_at_bottom_edge,
        mg.nodes_at_left_edge,
        mg.nodes_at_right_edge,
    ):
        mg.status_at_node[edge] = mg.BC_NODE_IS_CLOSED
    for edge in mg.nodes_at_bottom_edge:
        mg.status_at_node[edge] = mg.BC_NODE_IS_FIXED_VALUE

    z = mg.add_zeros("topographic__elevation", at="node")
    ir2 = np.random.uniform(low=0.0, high=0.5, size=(z.size))
    loading_vector = np.linspace(1, 2.5, num=nr)
    ramp = np.repeat(loading_vector, nc)
    ramp += ir2
    z[:] += ramp
    fa = FlowAccumulator(
        mg,
        surface="topographic__elevation",
        flow_director="FlowDirectorD8",
        runoff_rate=None,
        depression_finder=None,
    )
    latero = LateralEroder(
        mg, latero_mech="UC", alph=1.5, Kv=0.001, Kl_ratio=1.0, solver="basic"
    )

    for i in range(2000):
        fa.run_one_step()  # flow accumulator
        (mg, dzlat) = latero.run_one_step(dt)
        mg.at_node["topographic__elevation"][mg.core_nodes] += (
            U * dt
        )  # uplift the landscape

    # compare numerical and analytical sediment flux solutions
    num_sedflux = mg.at_node["qs"][mg.core_nodes]
    analytical_sedflux = U * mg.at_node["drainage_area"][mg.core_nodes]

    # test for match with analytical sediment flux.
    testing.assert_array_almost_equal(
        num_sedflux,
        analytical_sedflux,
        decimal=2,
        err_msg="LatEro transport-limited sediment flux test failed",
        verbose=True,
    )


def test_variable_bedrock_K():
    """
    Test that model matches the detachment-limited analytical solution
    for slope/area relationship at steady state: S=(U/K_br)^(1/n)*A^(-m/n),
    with variable bedrock erodibility.

    """
    U = 0.005  # m/yr
    dt = 1  # years

    nr = 5
    nc = 5
    nnodes = nr * nc
    dx = 10
    # instantiate grid
    mg = RasterModelGrid((nr, nc), xy_spacing=dx)
    for edge in (
        mg.nodes_at_top_edge,
        mg.nodes_at_bottom_edge,
        mg.nodes_at_left_edge,
        mg.nodes_at_right_edge,
    ):
        mg.status_at_node[edge] = mg.BC_NODE_IS_CLOSED
    for edge in mg.nodes_at_bottom_edge:
        mg.status_at_node[edge] = mg.BC_NODE_IS_FIXED_VALUE

    z = mg.add_zeros("topographic__elevation", at="node")
    loading_vector = np.linspace(1, 4, num=nr)
    ramp = np.repeat(loading_vector, nc)
    ramp += np.random.random_sample(nnodes) * 0.8
    z[:] += ramp
    n_sp = 1.0
    m_sp = 0.5
    # below, array of different bedrock erodibilities
    Kvar = 0.006 * np.ones(nr * nc)
    Kvar[0:9] = 0.06
    fa = FlowAccumulator(
        mg,
        surface="topographic__elevation",
        flow_director="FlowDirectorD8",
        runoff_rate=None,
        depression_finder=None,
    )  # "DepressionFinderAndRouter", router="D8")

    # ***NOTE, YOU MUST USE ADAPTIVE TIME STEPPER FOR variable K, or you may get strange
    # topography
    latero = LateralEroder(
        mg,
        latero_mech="UC",
        Kv=Kvar,
        solver="adaptive",
        alph=0.0,
        Kl_ratio=0.0,
        flow_accumulator=fa,
    )

    for i in range(2000):
        fa.run_one_step()  # flow accumulator
        # erode the landscape with lateral erosion
        (mg, dzlat) = latero.run_one_step(dt)
        mg.at_node["topographic__elevation"][mg.core_nodes] += U * dt

    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(U / Kvar[mg.core_nodes], 1.0 / n_sp) * np.power(
        mg.at_node["drainage_area"][mg.core_nodes], -m_sp / n_sp
    )

    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=4,
        err_msg="LatEro variable bedrock test failed",
        verbose=True,
    )


def test_latero_steady_inlet():
    """
    Test that time steady inputs of drainage area and sediment at a designated inlet node
    functions in the lateral erosion component.
    """
    U = 0.005  # m/yr
    dt = 10  # years
    nr = 5
    nc = 5
    dx = 10
    # instantiate grid
    mg = RasterModelGrid((nr, nc), xy_spacing=dx)
    for edge in (
        mg.nodes_at_top_edge,
        mg.nodes_at_bottom_edge,
        mg.nodes_at_left_edge,
        mg.nodes_at_right_edge,
    ):
        mg.status_at_node[edge] = mg.BC_NODE_IS_CLOSED
    for edge in mg.nodes_at_bottom_edge:
        mg.status_at_node[edge] = mg.BC_NODE_IS_FIXED_VALUE

    z = mg.add_zeros("topographic__elevation", at="node")
    loading_vector = np.linspace(1, 2.5, num=nr)
    ramp = np.repeat(loading_vector, nc)
    ramp += np.random.random_sample(mg.number_of_nodes) * 0.8
    z += ramp

    fa = FlowAccumulator(
        mg,
        surface="topographic__elevation",
        flow_director="FlowDirectorD8",
        runoff_rate=None,
        depression_finder=None,
    )
    # set inlet node = true, provide inlet node id, inlet drainage area, and inlet qs.
    latero = LateralEroder(
        mg,
        latero_mech="UC",
        Kv=0.001,
        Kl_ratio=1.0,
        inlet_on=True,
        inlet_node=17,
        inlet_area=500,
        qsinlet=2.5,
    )

    for i in range(2000):
        fa.run_one_step()  # flow accumulator
        # erode the landscape with lateral erosion
        (mg, dzlat) = latero.run_one_step(dt)
        mg.at_node["topographic__elevation"][mg.core_nodes] += (
            U * dt
        )  # uplift the landscape

    da = mg.at_node["surface_water__discharge"] / dx ** 2
    num_sedflux = mg.at_node["qs"]
    analytical_sedflux = U * da

    # test for match with analytical sediment flux. note that the values are off a little
    # because of the lateral erosion
    testing.assert_array_almost_equal(
        num_sedflux,
        analytical_sedflux,
        decimal=2,
        err_msg="LatEro inlet transport-limited sediment flux test failed",
        verbose=True,
    )



r"""
Examples
--------

::

        * - *
       / \ / \
      * - * - *
     / \ / \ / \
    * - * - * - *
     \ / \ / \ /
      * - * - *
       \ / \ /
        * - *

>>> from landlab.graph import TriGraph
>>> graph = TriGraph((5, 2), node_layout="hex", sort=True)
>>> graph.number_of_nodes
14
>>> graph.x_of_node
array([ 1. ,  2. ,
        0.5,  1.5,  2.5,
        0. ,  1. ,  2. ,  3. ,
        0.5,  1.5, 2.5,
        1. ,  2. ])
>>> graph.number_of_links
29
>>> graph.number_of_patches
16

::

    * - * - * - *
     \ / \ / \ / \
      * - * - * - *
     / \ / \ / \ /
    * - * - * - *

>>> from landlab.graph import TriGraph
>>> graph = TriGraph((3, 4), orientation="horizontal", node_layout="rect", sort=True)
>>> graph.number_of_nodes
12
>>> graph.x_of_node.reshape((3, 4))
array([[ 0. ,  1. ,  2. ,  3. ],
       [ 0.5,  1.5,  2.5,  3.5],
       [ 0. ,  1. ,  2. ,  3. ]])
>>> graph.number_of_links
23
>>> graph.number_of_patches
12
"""

from functools import lru_cache

import numpy as np

from ...core.utils import as_id_array
from ...utils.decorators import cache_result_in_object, make_return_array_immutable
from ..graph import Graph
from ..voronoi.voronoi import DelaunayGraph


class HorizontalRectTriGraphCython:
    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_lower_left=(0.0, 0.0)):
        from .ext.hex import fill_xy_of_node_rect_horizontal

        x_of_node = np.empty(shape[0] * shape[1], dtype=float)
        y_of_node = np.empty(shape[0] * shape[1], dtype=float)
        fill_xy_of_node_rect_horizontal(shape, x_of_node, y_of_node)

        x_of_node[:] *= spacing
        y_of_node[:] *= spacing * np.sin(np.pi / 3.0)
        x_of_node[:] += xy_of_lower_left[0]
        y_of_node[:] += xy_of_lower_left[1]

        return x_of_node, y_of_node


class VerticalRectTriGraphCython:
    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_lower_left=(0.0, 0.0)):
        from .ext.hex import fill_xy_of_node_rect_vertical

        n_rows, n_cols = shape
        x_spacing = np.sin(np.pi / 3.0) * spacing
        y_spacing = spacing

        x_of_node = np.empty(n_rows * n_cols, dtype=float)
        y_of_node = np.empty(n_rows * n_cols, dtype=float)

        fill_xy_of_node_rect_vertical(shape, x_of_node, y_of_node)

        x_of_node *= x_spacing
        y_of_node *= y_spacing
        x_of_node += xy_of_lower_left[0]
        y_of_node += xy_of_lower_left[1]

        return x_of_node, y_of_node

        x_of_node[:, (n_cols + 1) // 2 :] = (
            np.arange(n_cols // 2) * x_spacing * 2.0 + x_spacing + xy_of_lower_left[0]
        )
        x_of_node[:, : (n_cols + 1) // 2] = (
            np.arange((n_cols + 1) // 2) * x_spacing * 2.0 + xy_of_lower_left[0]
        )

        y_of_node[:, : (n_cols + 1) // 2] = (
            np.arange(n_rows) * y_spacing + xy_of_lower_left[1]
        ).reshape((n_rows, 1))
        y_of_node[:, (n_cols + 1) // 2 :] = (
            np.arange(n_rows) * y_spacing + xy_of_lower_left[1] + y_spacing * 0.5
        ).reshape((n_rows, 1))

        return x_of_node.reshape(-1), y_of_node.reshape(-1)


class HorizontalHexTriGraphCython:
    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_lower_left=(0.0, 0.0)):
        from .ext.hex import fill_xy_of_node_hex_horizontal

        n_rows, n_cols = shape
        n_nodes = n_rows * n_cols + (n_rows // 2) ** 2

        x_of_node = np.empty(n_nodes, dtype=float)
        y_of_node = np.empty(n_nodes, dtype=float)

        fill_xy_of_node_hex_horizontal(shape, x_of_node, y_of_node)

        x_of_node[:] *= spacing
        y_of_node[:] *= spacing * np.sin(np.pi / 3.0)
        x_of_node[:] += xy_of_lower_left[0]
        y_of_node[:] += xy_of_lower_left[1]

        return x_of_node, y_of_node


class VerticalHexTriGraphCython:
    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_lower_left=(0.0, 0.0)):
        from .ext.hex import fill_xy_of_node_hex_vertical

        n_rows, n_cols = shape
        n_nodes = n_cols * n_rows + (n_cols // 2) ** 2

        x_of_node = np.empty(n_nodes, dtype=float)
        y_of_node = np.empty(n_nodes, dtype=float)

        fill_xy_of_node_hex_vertical(shape, x_of_node, y_of_node)

        x_of_node[:] *= spacing * np.sin(np.pi / 3.0)
        y_of_node[:] *= spacing
        x_of_node[:] += xy_of_lower_left[0]
        y_of_node[:] += xy_of_lower_left[1]

        return x_of_node, y_of_node


class HorizontalRectTriGraph:
    @staticmethod
    def number_of_nodes(shape):
        n_rows, n_cols = shape
        return n_rows * n_cols

    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_lower_left=(0.0, 0.0)):
        n_rows, n_cols = shape
        x_spacing, y_spacing = spacing, spacing * np.sin(np.pi / 3.0)

        x_of_node, y_of_node = np.meshgrid(
            np.arange(n_cols) * x_spacing + xy_of_lower_left[0],
            np.arange(n_rows) * y_spacing + xy_of_lower_left[1],
        )
        x_of_node[1::2] += spacing * 0.5

        return x_of_node.reshape(-1), y_of_node.reshape(-1)

    @staticmethod
    def corner_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import HorizontalRectTriGraph
        >>> HorizontalRectTriGraph.corner_nodes((3, 4))
        (11, 8, 0, 3)
        >>> HorizontalRectTriGraph.corner_nodes((3, 2))
        (5, 4, 0, 1)
        >>> HorizontalRectTriGraph.corner_nodes((7, 1))
        (6, 6, 0, 0)
        >>> HorizontalRectTriGraph.corner_nodes((1, 3))
        (2, 0, 0, 2)
        """
        n_rows, n_cols = shape
        return (n_rows * n_cols - 1, n_cols * (n_rows - 1), 0, n_cols - 1)

    @staticmethod
    def number_of_perimeter_nodes(shape):
        if 1 in shape:
            return np.prod(shape)
        return 2 * shape[0] + 2 * (shape[1] - 2)

    @staticmethod
    def perimeter_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import HorizontalRectTriGraph
        >>> HorizontalRectTriGraph.perimeter_nodes((3, 2))
        array([1, 3, 5, 4, 2, 0])
        """
        return np.concatenate(HorizontalRectTriGraph.nodes_at_edge(shape))

    @staticmethod
    def nodes_at_edge(shape):
        n_rows, n_cols = shape
        if n_rows == n_cols == 1:
            return (np.array([0]),) + (np.array([], dtype=int),) * 3
        (
            northeast,
            northwest,
            southwest,
            southeast,
        ) = HorizontalRectTriGraph.corner_nodes(shape)

        if n_rows > 1:
            south = np.arange(southwest, southeast)
        else:
            south = np.array([southwest], dtype=int)

        if n_cols > 1:
            west = np.arange(northwest, southwest, -n_cols)
        else:
            west = np.array([northwest], dtype=int)

        return (
            np.arange(southeast, northeast, n_cols),
            np.arange(northeast, northwest, -1),
            west,
            south,
        )


class VerticalRectTriGraph:
    @staticmethod
    def number_of_nodes(shape):
        n_rows, n_cols = shape
        return n_rows * n_cols

    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_lower_left=(0.0, 0.0)):
        n_rows, n_cols = shape
        x_spacing, y_spacing = spacing * np.sin(np.pi / 3.0), spacing

        x_of_node = np.empty((n_rows, n_cols), dtype=float)
        y_of_node = np.empty((n_rows, n_cols), dtype=float)

        x_of_node[:, (n_cols + 1) // 2 :] = (
            np.arange(n_cols // 2) * x_spacing * 2.0 + x_spacing + xy_of_lower_left[0]
        )
        x_of_node[:, : (n_cols + 1) // 2] = (
            np.arange((n_cols + 1) // 2) * x_spacing * 2.0 + xy_of_lower_left[0]
        )

        y_of_node[:, : (n_cols + 1) // 2] = (
            np.arange(n_rows) * y_spacing + xy_of_lower_left[1]
        ).reshape((n_rows, 1))
        y_of_node[:, (n_cols + 1) // 2 :] = (
            np.arange(n_rows) * y_spacing + xy_of_lower_left[1] + y_spacing * 0.5
        ).reshape((n_rows, 1))

        return x_of_node.reshape(-1), y_of_node.reshape(-1)

    @staticmethod
    def corner_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import VerticalRectTriGraph
        >>> VerticalRectTriGraph.corner_nodes((4, 3))
        (10, 9, 0, 1)
        >>> VerticalRectTriGraph.corner_nodes((4, 4))
        (15, 12, 0, 3)
        >>> VerticalRectTriGraph.corner_nodes((3, 2))
        (5, 4, 0, 1)
        >>> VerticalRectTriGraph.corner_nodes((7, 1))
        (6, 6, 0, 0)
        >>> VerticalRectTriGraph.corner_nodes((1, 3))
        (1, 0, 0, 1)
        >>> VerticalRectTriGraph.corner_nodes((2, 3))
        (4, 3, 0, 1)
        """
        n_rows, n_cols = shape
        if n_cols % 2 == 0:
            return (n_rows * n_cols - 1, n_cols * (n_rows - 1), 0, n_cols - 1)
        else:
            return (
                n_rows * n_cols - 1 - n_cols // 2,
                n_cols * (n_rows - 1),
                0,
                n_cols // 2,
            )

    @staticmethod
    def number_of_perimeter_nodes(shape):
        if 1 in shape:
            return np.prod(shape)
        return 2 * shape[1] + 2 * (shape[0] - 2)

    @staticmethod
    def perimeter_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import VerticalRectTriGraph
        >>> VerticalRectTriGraph.perimeter_nodes((3, 2))
        array([1, 3, 5, 4, 2, 0])
        >>> VerticalRectTriGraph.perimeter_nodes((2, 3))
        array([1, 4, 5, 3, 0, 2])
        >>> VerticalRectTriGraph.perimeter_nodes((2, 4))
        array([3, 7, 5, 6, 4, 0, 2, 1])
        """
        return np.concatenate(VerticalRectTriGraph.nodes_at_edge(shape))

    @staticmethod
    def nodes_at_edge(shape):
        n_rows, n_cols = shape
        if n_rows == n_cols == 1:
            return (np.array([0]),) + (np.array([], dtype=int),) * 3
        n_nodes = n_rows * n_cols
        (
            northeast,
            northwest,
            southwest,
            southeast,
        ) = VerticalRectTriGraph.corner_nodes(shape)

        if n_cols == 1:
            southwest = northwest - n_cols

        north = np.empty(n_cols - 1, dtype=int)
        north[::2] = n_nodes - n_cols // 2 + np.arange(n_cols // 2)
        north[1::2] = northwest + np.arange(1, n_cols - n_cols // 2)

        if n_rows > 1:
            south = np.empty(n_cols - 1, dtype=int)
            south[::2] = np.arange(0, n_cols // 2)
            south[1::2] = (n_cols + 1) // 2 + np.arange(n_cols - n_cols // 2 - 1)
        else:
            south = np.array([southwest], dtype=int)

        return (
            np.arange(southeast, northeast, n_cols),
            north[::-1],
            np.arange(northwest, southwest, -n_cols),
            south,
        )


class HorizontalHexTriGraph:
    @staticmethod
    def number_of_nodes(shape):
        n_rows, n_cols = shape
        return n_rows * n_cols + (n_rows // 2) ** 2

    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_lower_left=(0.0, 0.0)):
        n_rows, n_cols = shape
        x_spacing, y_spacing = spacing, spacing * np.sin(np.pi / 3.0)

        length_of_row = np.concatenate(
            (
                np.arange((n_rows + 2) // 2) + n_cols,
                (n_rows + 2) // 2
                + n_cols
                - 1
                - np.arange(1, n_rows - (n_rows + 2) // 2 + 1),
            )
        )
        offset_to_row = np.concatenate((np.array([0]), length_of_row)).cumsum()
        rows = [
            slice(start, end)
            for start, end in zip(offset_to_row[:-1], offset_to_row[1:])
        ]

        y_of_node = np.empty(HorizontalHexTriGraph.number_of_nodes(shape), dtype=float)
        for row, inds in enumerate(rows):
            y_of_node[inds] = row * y_spacing + xy_of_lower_left[1]

        x_of_node = np.empty(HorizontalHexTriGraph.number_of_nodes(shape), dtype=float)

        x_of_row = (
            np.abs((n_rows + 2) // 2 - 1 - np.arange(n_rows)) * x_spacing * 0.5
            + xy_of_lower_left[0]
        )
        for row, inds in enumerate(rows):
            x_of_node[inds] = x_of_row[row] + np.arange(length_of_row[row]) * x_spacing

        return x_of_node.reshape(-1), y_of_node.reshape(-1)

    @staticmethod
    def corner_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import HorizontalHexTriGraph
        >>> HorizontalHexTriGraph.corner_nodes((3, 2))
        (4, 6, 5, 2, 0, 1)
        >>> HorizontalHexTriGraph.corner_nodes((7, 1))
        (9, 15, 15, 6, 0, 0)
        >>> HorizontalHexTriGraph.corner_nodes((6, 1))
        (9, 14, 13, 6, 0, 0)
        >>> HorizontalHexTriGraph.corner_nodes((4, 2))
        (8, 11, 9, 5, 0, 1)
        """
        n_rows, n_cols = shape

        n_nodes_in_middle_row = n_rows // 2 + n_cols
        n_nodes = n_rows * n_cols + (n_rows // 2) ** 2

        east = (n_nodes_in_middle_row + n_cols) * ((n_rows // 2 + 1) // 2) - 1
        if (n_rows // 2) % 2 == 0:
            east += (n_nodes_in_middle_row + n_cols) // 2

        return (
            east,
            n_nodes - 1,
            n_nodes - (n_cols + (n_rows + 1) % 2),
            east - (n_nodes_in_middle_row - 1),
            0,
            n_cols - 1,
        )

    @staticmethod
    def number_of_perimeter_nodes(shape):
        if shape[0] == 1:
            return shape[1]
        return 2 * shape[0] + 2 * (shape[1] - 2) + (shape[0] + 1) % 2

    @staticmethod
    def perimeter_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import HorizontalHexTriGraph
        >>> HorizontalHexTriGraph.perimeter_nodes((3, 2))
        array([4, 6, 5, 2, 0, 1])
        >>> HorizontalHexTriGraph.perimeter_nodes((1, 3))
        array([2, 1, 0])
        """
        return np.concatenate(HorizontalHexTriGraph.nodes_at_edge(shape))

    @staticmethod
    def nodes_at_edge(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import HorizontalHexTriGraph
        >>> HorizontalHexTriGraph.nodes_at_edge((5, 3))
        (array([11, 15]), array([18, 17]), array([16, 12]), array([7, 3]), array([0, 1]), array([2, 6]))
        >>> HorizontalHexTriGraph.nodes_at_edge((4, 3))
        (array([11]), array([15, 14, 13]), array([12]), array([7, 3]), array([0, 1]), array([2, 6]))
        """
        n_rows, n_cols = shape
        (
            east,
            northeast,
            northwest,
            west,
            southwest,
            southeast,
        ) = HorizontalHexTriGraph.corner_nodes(shape)

        if n_rows == 1:
            nodes_at_south_edge = np.asarray([southwest], dtype=int)
        else:
            nodes_at_south_edge = np.arange(southwest, southeast)

        return (
            as_id_array(
                northeast
                - np.arange(northeast - northwest + 1, east - west + 1).cumsum()[::-1]
            ),
            np.arange(northeast, northwest, -1),
            as_id_array(
                west
                + np.arange(east - west + 1, northeast - northwest + 1, -1).cumsum()[
                    ::-1
                ]
            ),
            as_id_array(
                southwest + np.arange(n_cols, n_rows // 2 + n_cols).cumsum()[::-1]
            ),
            nodes_at_south_edge,
            as_id_array(
                east - np.arange(n_cols + n_rows // 2, n_cols, -1).cumsum()[::-1]
            ),
        )


class VerticalHexTriGraph:
    @staticmethod
    def number_of_nodes(shape):
        n_rows, n_cols = shape
        return n_rows * n_cols + (n_cols // 2) ** 2

    @staticmethod
    def xy_of_node(shape, spacing=1.0, xy_of_lower_left=(0.0, 0.0)):
        n_rows, n_cols = shape
        x_spacing, y_spacing = spacing * np.sin(np.pi / 3.0), spacing

        length_of_middle_rows = np.full(2 * n_rows - 1, n_cols // 2)
        if n_cols % 2 == 1:
            length_of_middle_rows[::2] += 1

        length_of_row = np.concatenate(
            (
                np.arange(1, n_cols // 2 + 1),
                length_of_middle_rows,
                np.arange(n_cols // 2, 0, -1),
            )
        )
        offset_to_row = np.concatenate((np.array([0]), length_of_row)).cumsum()
        rows = [
            slice(start, end)
            for start, end in zip(offset_to_row[:-1], offset_to_row[1:])
        ]

        y_of_node = np.empty(VerticalHexTriGraph.number_of_nodes(shape), dtype=float)
        for row, inds in enumerate(rows):
            y_of_node[inds] = row * y_spacing * 0.5

        x_of_node = np.empty(VerticalHexTriGraph.number_of_nodes(shape), dtype=float)

        x_of_middle_rows = np.zeros(2 * n_rows - 1)
        x_of_middle_rows[1::2] += 1.0
        x_of_row = (
            np.concatenate(
                (
                    np.arange(n_cols // 2, 0, -1),
                    x_of_middle_rows,
                    np.arange(1, n_cols // 2 + 1),
                )
            )
            * x_spacing
        )
        for row, inds in enumerate(rows):
            x_of_node[inds] = (
                x_of_row[row] + np.arange(length_of_row[row]) * 2.0 * x_spacing
            )

        x_of_node += xy_of_lower_left[0]
        y_of_node += xy_of_lower_left[1]

        return x_of_node.reshape(-1), y_of_node.reshape(-1)

    @staticmethod
    def corner_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import VerticalHexTriGraph
        >>> VerticalHexTriGraph.corner_nodes((2, 5))
        (10, 13, 8, 3, 0, 5)
        >>> VerticalHexTriGraph.corner_nodes((2, 3))
        (5, 6, 4, 1, 0, 2)
        >>> VerticalHexTriGraph.corner_nodes((2, 4))
        (10, 11, 7, 3, 0, 2)
        >>> VerticalHexTriGraph.corner_nodes((2, 2))
        (4, 4, 3, 1, 0, 0)
        >>> VerticalHexTriGraph.corner_nodes((3, 1))
        (2, 2, 2, 0, 0, 0)
        >>> VerticalHexTriGraph.corner_nodes((1, 3))
        (2, 3, 1, 1, 0, 2)
        """
        n_rows, n_cols = shape
        n_nodes = n_rows * n_cols + (n_cols // 2) ** 2

        n = n_cols // 2
        tri_nodes = (n + 1) * (n // 2)
        if n % 2 == 1:
            tri_nodes += (n + 1) // 2

        if n_cols % 2 == 0:
            southeast = tri_nodes - 1
            southwest = tri_nodes
            northwest = n_nodes - 1 - (tri_nodes - 1 + n_cols // 2)
            northeast = n_nodes - 1 - (tri_nodes - n)

        else:
            southwest = tri_nodes
            southeast = tri_nodes + n_cols // 2
            northwest = n_nodes - (tri_nodes + (n_cols + 1) // 2)
            northeast = n_nodes - 1 - tri_nodes

        south = 0
        north = n_nodes - 1

        return (northeast, north, northwest, southwest, south, southeast)

    @staticmethod
    def number_of_perimeter_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import VerticalHexTriGraph
        >>> VerticalHexTriGraph.number_of_perimeter_nodes((2, 3))
        6
        >>> VerticalHexTriGraph.number_of_perimeter_nodes((2, 2))
        5
        """
        if shape[1] == 1:
            return shape[0]
        return 2 * shape[1] + 2 * (shape[0] - 2) + (shape[1] + 1) % 2

    @staticmethod
    def perimeter_nodes(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import VerticalHexTriGraph
        >>> VerticalHexTriGraph.perimeter_nodes((3, 7))
        array([ 9, 16, 23, 26, 28, 29, 27, 24, 20, 13,  6,  3,  1,  0,  2,  5])
        >>> VerticalHexTriGraph.perimeter_nodes((2, 3))
        array([2, 5, 6, 4, 1, 0])
        >>> VerticalHexTriGraph.perimeter_nodes((2, 4))
        array([ 2, 6, 10, 11, 9, 7, 3, 1, 0])
        >>> VerticalHexTriGraph.perimeter_nodes((2, 2))
        array([0, 2, 4, 3, 1])
        >>> VerticalHexTriGraph.perimeter_nodes((3, 1))
        array([0, 1, 2])
        """
        return np.concatenate(VerticalHexTriGraph.nodes_at_edge(shape))

    @staticmethod
    def nodes_at_edge(shape):
        """
        Examples
        --------
        >>> from landlab.graph.hex.hex import VerticalHexTriGraph
        >>> VerticalHexTriGraph.nodes_at_edge((3, 7))
        (array([ 9, 16]), array([23, 26, 28]), array([29, 27, 24]), array([20, 13]), array([6, 3, 1]), array([0, 2, 5]))
        >>> VerticalHexTriGraph.nodes_at_edge((2, 3))
        (array([2]), array([5]), array([6]), array([4]), array([1]), array([0]))
        >>> VerticalHexTriGraph.nodes_at_edge((2, 4))
        (array([2, 6]), array([10]), array([11, 9]), array([7]), array([3, 1]), array([0]))
        """
        n_rows, n_cols = shape
        (
            northeast,
            north,
            northwest,
            southwest,
            south,
            southeast,
        ) = VerticalHexTriGraph.corner_nodes(shape)

        if shape[1] == 1:
            southwest = northwest - n_cols

        return (
            np.arange(southeast, northeast, n_cols),
            as_id_array(north - np.arange(1, (n_cols + 1) // 2).cumsum())[::-1],
            as_id_array(north - np.arange(1, (n_cols + 2) // 2).cumsum() + 1),
            np.arange(northwest, southwest, -n_cols),
            as_id_array(south + np.arange(1, (n_cols + 2) // 2).cumsum())[::-1],
            as_id_array(south + np.arange(1, (n_cols + 1) // 2).cumsum() - 1),
        )


class HexGraphExtras:
    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def nodes_at_right_edge(self):
        """Get nodes along the right edge.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph import TriGraph
        >>> graph = TriGraph((3, 4), node_layout='rect')
        >>> graph.nodes_at_right_edge
        array([ 3,  7, 11])
        """
        return np.arange(
            self.shape[1] - 1, self.shape[0] * self.shape[1], self.shape[1], dtype=int
        )

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def nodes_at_top_edge(self):
        """Get nodes along the top edge.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph import TriGraph
        >>> graph = TriGraph((3, 4), node_layout='rect')
        >>> graph.nodes_at_top_edge
        array([ 8,  9, 10, 11])
        """
        return np.arange(
            self.number_of_nodes - self.shape[1], self.number_of_nodes, dtype=int
        )

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def nodes_at_left_edge(self):
        """Get nodes along the left edge.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph import TriGraph
        >>> graph = TriGraph((3, 4), node_layout='rect')
        >>> graph.nodes_at_left_edge
        array([0, 4, 8])
        """
        return np.arange(0, self.shape[0] * self.shape[1], self.shape[1], dtype=int)

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def nodes_at_bottom_edge(self):
        """Get nodes along the bottom edge.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph import TriGraph
        >>> graph = TriGraph((3, 4), node_layout='rect')
        >>> graph.nodes_at_bottom_edge
        array([0, 1, 2, 3])
        """
        return np.arange(self.shape[1], dtype=int)

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def length_of_link(self):
        return np.full(self.number_of_links, self.spacing, dtype=float)


class TriGraph(HexGraphExtras, DelaunayGraph):

    """Graph of a structured grid of triangles.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import TriGraph

    >>> graph = TriGraph((3, 2))
    >>> graph.number_of_nodes == 6
    True
    >>> np.round(graph.y_of_node * 2. / np.sqrt(3))
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([ 0.,  0.,  1.,  1.,  2.,  2.])
    >>> graph.x_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 0. ,  1. ,  0.5,  1.5,  0. ,  1. ])
    """

    def __init__(
        self,
        shape,
        spacing=1.0,
        xy_of_lower_left=(0.0, 0.0),
        orientation="horizontal",
        node_layout="rect",
        sort=False,
    ):
        """Create a structured grid of triangles.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of the hex grid. The first value
            is the number of nodes in the first column and the second the
            number of nodes in the first column.
        spacing : float, optional
            Length of links.
        xy_of_lower_left : tuple of float, optional
            Coordinates of lower-left corner of the grid.
        orientation: {'horizontal', 'vertical'}
            Specify if triangles should be laid out in rows or columns.
        node_layout: {'rect', 'hex'}
            Specify the overall layout of the nodes. Use *rect* for
            the layout to approximate a rectangle and *hex* for
            a hexagon.
        """
        if node_layout not in ("rect", "hex"):
            raise ValueError("node_layout not understood")

        if orientation not in ("horizontal", "vertical"):
            raise ValueError("orientation not understood")

        layouts = {
            "horizontal_hex": HorizontalHexTriGraph,
            "vertical_hex": VerticalHexTriGraph,
            "horizontal_rect": HorizontalRectTriGraph,
            "vertical_rect": VerticalRectTriGraph,
        }
        layout = layouts["_".join([orientation, node_layout])]

        try:
            spacing = float(spacing)
        except TypeError:
            raise TypeError("spacing must be a float")

        self._shape = tuple(shape)
        self._spacing = spacing
        self._orientation = orientation
        self._node_layout = node_layout

        x_of_node, y_of_node = layout.xy_of_node(
            shape, spacing=spacing, xy_of_lower_left=xy_of_lower_left
        )
        self._perimeter_nodes = layout.perimeter_nodes(shape)

        perimeter_links = np.empty((len(self._perimeter_nodes), 2), dtype=int)
        perimeter_links[:, 0] = self._perimeter_nodes
        perimeter_links[:-1, 1] = self._perimeter_nodes[1:]
        perimeter_links[-1, 1] = self._perimeter_nodes[0]

        if 1 in shape:
            Graph.__init__(
                self,
                (y_of_node, x_of_node),
                links=list(
                    zip(np.arange(len(y_of_node) - 1), np.arange(1, len(y_of_node)))
                ),
                sort=False,
            )
        else:
            DelaunayGraph.__init__(
                self,
                (y_of_node, x_of_node),
                perimeter_links=perimeter_links,
                sort=False,
            )

        if sort:
            self.sort()

    @property
    def shape(self):
        return self._shape

    @property
    def spacing(self):
        return self._spacing

    @property
    def orientation(self):
        return self._orientation

    @property
    def node_layout(self):
        return self._node_layout

    @property
    @lru_cache()
    @make_return_array_immutable
    def perimeter_nodes(self):
        return self._perimeter_nodes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Created on Tue Apr 11 10:13:38 2017.

@author: margauxmouchene
"""


import numpy as np

from landlab import Component


class TransportLengthHillslopeDiffuser(Component):

    r"""Transport length hillslope diffusion.

    Hillslope diffusion component in the style of Carretier et al. (2016,
    ESurf), and Davy and Lague (2009)

    .. math::

        \frac{dz}{dt} = -E + D (+ U)

        D = \frac{q_s}{L}

        E = k S

        L = \frac{dx}{(1 - (S / S_c)^2}

    Works on regular raster-type grid (RasterModelGrid, dx=dy).
    To be coupled with FlowDirectorSteepest for the calculation of steepest
    slope at each timestep.

    Component written by Margaux Mouchene, 2017

    Parameters
    ----------
    grid : ModelGrid
        Landlab ModelGrid object
    erodibility: float
        Erodibility coefficient [L/T]
    slope_crit: float (default=1.)
        Critical slope [L/L]

    Examples
    --------

    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowDirectorSteepest
    >>> from landlab.components import TransportLengthHillslopeDiffuser

    Define grid and initial topography:

        - 3x5 grid
        - east and west boundaries are open, north and south are closed
        - Initial topography is plane at base level on the boundaries and
          1m of elevation elsewhere (core)

    >>> mg = RasterModelGrid((5, 5))
    >>> mg.set_closed_boundaries_at_grid_edges(False, True, False, True)
    >>> z = np.array([0., 0., 0., 0., 0.,
    ...               0., 1., 1., 1., 0.,
    ...               0., 1., 1., 1., 0.,
    ...               0., 1., 1., 1., 0.,
    ...               0., 0., 0., 0., 0.])
    >>> _ = mg.add_field("topographic__elevation", z, at="node")

    Instantiate Flow director (steepest slope type) and TL hillslope diffuser

    >>> fdir = FlowDirectorSteepest(mg)
    >>> tl_diff = TransportLengthHillslopeDiffuser(
    ...     mg,
    ...     erodibility=0.001,
    ...     slope_crit=0.6)

    Run the components for ten short timepsteps

    >>> for t in range(10):
    ...     fdir.run_one_step()
    ...     tl_diff.run_one_step(1.)

    Check final topography

    >>> np.allclose(
    ...     mg.at_node['topographic__elevation'],
    ...     np.array([ 0.,  0.        ,  0.        ,  0.        ,  0.,
    ...                0.,  0.96175283,  0.99982519,  0.96175283,  0.,
    ...                0.,  0.96175283,  0.99982519,  0.96175283,  0.,
    ...                0.,  0.96175283,  0.99982519,  0.96175283,  0.,
    ...                0.,  0.        ,  0.        ,  0.        ,  0.]))
    True

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    Carretier, S., Martinod, P., Reich, M., Godderis, Y. (2016). Modelling
    sediment clasts transport during landscape evolution. Earth Surface Dynamics
    4(1), 237-251. https://dx.doi.org/10.5194/esurf-4-237-2016

    Davy, P., Lague, D. (2009). Fluvial erosion/transport equation of landscape
    evolution models revisited. Journal of Geophysical Research  114(F3),
    F03007. https://dx.doi.org/10.1029/2008jf001146

    """

    _name = "TransportLengthHillslopeDiffuser"

    _unit_agnostic = True

    _info = {
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "sediment__deposition_coeff": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Fraction of incoming sediment that is deposited on the node",
        },
        "sediment__deposition_rate": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/yr",
            "mapping": "node",
            "doc": "Deposition rate on node",
        },
        "sediment__erosion_rate": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/yr",
            "mapping": "node",
            "doc": "Erosion rate on node",
        },
        "sediment__flux_in": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/yr",
            "mapping": "node",
            "doc": "Incoming sediment rate on node (=qs/dx)",
        },
        "sediment__flux_out": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/yr",
            "mapping": "node",
            "doc": "Outgoing sediment rate on node = sediment eroded on node + sediment transported across node from upstream",
        },
        "sediment__transfer_rate": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/yr",
            "mapping": "node",
            "doc": "Rate of transferred sediment across a node (incoming sediment - deposited sediment on node)",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m/m",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
    }

    def __init__(self, grid, erodibility=0.001, slope_crit=1.0):

        """Initialize Diffuser.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        erodibility: float
            Erodibility coefficient [L/T]
        slope_crit: float (default=1.)
            Critical slope [L/L]
        """
        super().__init__(grid)

        if grid.at_node["flow__receiver_node"].size != grid.size("node"):
            msg = (
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that TransportLengthHillslopeDiffuser is compatible "
                "with route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )
            raise NotImplementedError(msg)

        # Store grid and parameters

        self._k = erodibility
        self._slope_crit = slope_crit

        # Create fields:
        # Elevation
        self._elev = self._grid.at_node["topographic__elevation"]

        # Downstream steepest slope at node:
        self._steepest = self._grid.at_node["topographic__steepest_slope"]
        # On each node, node ID of downstream receiver node
        # (on node (i), ID of node that receives flow from node (i)):
        self._receiver = self._grid.at_node["flow__receiver_node"]

        self.initialize_output_fields()
        # Deposition
        self._depo = self._grid.at_node["sediment__deposition_rate"]

        # Transferred sediments (crossing over node)
        self._trans = self._grid.at_node["sediment__transfer_rate"]

        # Transport coefficient
        self._d_coeff = self._grid.at_node["sediment__deposition_coeff"]

        # Flux in
        self._flux_in = self._grid.at_node["sediment__flux_in"]

        # Flux out
        self._flux_out = self._grid.at_node["sediment__flux_out"]

        # Erosion
        self._erosion = self._grid.at_node["sediment__erosion_rate"]

    def tldiffusion(self, dt):
        """Calculate hillslope diffusion for a time period 'dt'.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        dt: float (time)
            The imposed timestep.
        """

        # Reset erosion, depo, trans and flux_in to 0
        self._erosion[:] = 0.0
        self._depo[:] = 0.0
        self._trans[:] = 0.0
        self._flux_in[:] = 0.0

        dx = self._grid.dx
        cores = self._grid.core_nodes

        # Calculate influx rate on node i  = outflux of nodes
        # whose receiver is i
        for i in self._grid.core_nodes:
            self._flux_in[self._receiver[i]] += self._flux_out[i]

            # Calculate transport coefficient
            # When S ~ Scrit, d_coeff is set to "infinity", for stability and
            # so that there is no deposition
            if self._steepest[i] >= self._slope_crit:
                self._d_coeff[i] = 1000000000.0
            else:
                self._d_coeff[i] = 1 / (
                    1 - (np.power(((self._steepest[i]) / self._slope_crit), 2))
                )

        # Calculate deposition rate on node
        self._depo[cores] = self._flux_in[cores] / self._d_coeff[cores]

        # Calculate erosion rate on node (positive value)
        # If S > Scrit, erosion is simply set for the slope to return to Scrit
        # Otherwise, erosion is slope times erodibility coefficent
        for i in self._grid.core_nodes:
            if self._steepest[i] > self._slope_crit:
                self._erosion[i] = (
                    dx * (self._steepest[i] - self._slope_crit) / (100 * dt)
                )
            else:
                self._erosion[i] = self._k * self._steepest[i]

            # Update elevation
            self._elev[i] += (-self._erosion[i] + self._depo[i]) * dt

        # Calculate transfer rate over node
        self._trans[cores] = self._flux_in[cores] - self._depo[cores]

        # Calculate outflux rate
        self._flux_out[:] = self._erosion + self._trans

    def run_one_step(self, dt):
        """Advance one timestep.

        Advance transport length-model hillslope diffusion component
        by one time step of size dt and tests for timestep stability.

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.
        """
        self.tldiffusion(dt)

        # Test code stability for timestep dt
        # Raise unstability error if local slope is reversed by erosion
        # and deposition during a timestep dt
        elev_dif = self._elev - self._elev[self._receiver]
        s = elev_dif[np.where(self._grid.at_node["flow__sink_flag"] == 0)]
        if np.any(s < -1) is True:
            raise ValueError(
                "The component is unstable" " for such a large timestep " "on this grid"
            )
        else:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Zone functions and class of SpeciesEvolver."""
from collections import OrderedDict
from enum import IntEnum, unique

import numpy as np
from pandas import Series


@unique
class Connection(IntEnum):
    """Zone connection type.

    Connection type represents the connectivity of zones in two time steps. It
    is described with the pattern, x-to-y where x and y are the descriptive
    counts (none, one, or many) of zone(s) at the earlier and later time step,
    respectively. For example, a connection type of one-to-many is a zone in
    the earlier time step that spatially overlaps multiple zones in the later
    time step.
    """

    NONE_TO_NONE = 0
    NONE_TO_ONE = 1
    ONE_TO_NONE = 2
    ONE_TO_ONE = 3
    ONE_TO_MANY = 4
    MANY_TO_ONE = 5
    MANY_TO_MANY = 6


def _update_zones(grid, prior_zones, new_zones, record):
    """Resolve the spatial connectivity of zones across two time steps.

    This method iterates over each zone of the prior time step to identify the
    zones of the current time step it spatially intersects. This method updates
    the zone attribute, ``successors``. Successor zones are the new zones
    existing at the current time step that are the continuation of prior time
    step zones referred to as the predecessor zones.

    The type of connection between predecessor and successor zones is described
    ``Connection``. The `_conn_type` property of zones are set by this method.

    In the `many` connections, a rule determines which of the prior zones
    persist as the zone in the current time step. The zone with the greatest
    area of intersection between the prior and current time steps persists to
    the current time step along with the others in `new_zones`.

    Parameters
    ----------
    grid : ModelGrid
        A Landlab ModelGrid.
    prior_zones : Zone list
        The zones of the prior timestep.
    new_zones : Zone list
        The zones of the current timestep.
    record : Record
        The ZoneController record.

    Returns
    -------
    list of Zones
        The successor zones. The zones determined to exist at the current time
        step with an updated ``successors`` attribute.
    """
    # Stats are calculated for `record`.

    fragment_ct = 0
    capture_ct = 0
    area_captured = [0]

    # Get `successors`.

    if len(prior_zones) == 0 and len(new_zones) == 0:
        successors = []

    elif len(prior_zones) == 0 and len(new_zones) > 0:
        successors = new_zones
        conn_type = _determine_connection_type(0, 1)
        for n in new_zones:
            n._conn_type = conn_type
            n._successors = [n]

    elif len(prior_zones) > 0 and len(new_zones) == 0:
        successors = []
        conn_type = _determine_connection_type(1, 0)
        for p in prior_zones:
            p._conn_type = conn_type
            p._successors = []

    elif len(prior_zones) > 0 and len(new_zones) > 0:
        successors = []

        # Get index maps for the prior zones (`ps`) and new zones (`ns`).
        ps_index_map = _create_index_map(grid, prior_zones)
        ns_index_map = _create_index_map(grid, new_zones)

        replacements = OrderedDict()

        for i_p, p in enumerate(prior_zones):
            # Retain a copy of the prior zone mask to compare with the new zone
            # mask after the zone masks are updated.
            p_mask_copy = p.mask.copy()

            # Get the new zones that intersect (`i`) the prior zone.
            ns_i_p = _intersecting_zones(ps_index_map == i_p, ns_index_map, new_zones)
            ns_i_p_ct = len(ns_i_p)

            # Get the other prior zones that intersect the new zones.
            ns_mask = np.any([n.mask for n in ns_i_p], axis=0)
            ps_i_ns = _intersecting_zones(ns_mask, ps_index_map, prior_zones)
            ps_i_ns_ct = len(ps_i_ns)

            if ps_i_ns_ct == 0:
                ps_i_ns_ct = 1
                ps_i_ns = p

            # Get successors depending on connection type.

            conn_type = _determine_connection_type(ps_i_ns_ct, ns_i_p_ct)

            p_successors = _get_successors(
                p,
                conn_type,
                ps_i_ns,
                ns_i_p,
                prior_zones,
                ps_index_map,
                replacements,
                successors,
            )

            # Update statistics.

            if conn_type in [Connection.MANY_TO_ONE, Connection.MANY_TO_MANY]:
                # Copy successors to be `captured_zones`.
                captured_zones = list(p_successors)
                if p in captured_zones:
                    captured_zones.remove(p)
                capture_ct += len(captured_zones)

                for z in captured_zones:
                    captured_mask = np.all([~p_mask_copy, z.mask], 0)
                    area = grid.cell_area_at_node[captured_mask.flatten()].sum()
                    area_captured.append(area)

            elif conn_type in [Connection.ONE_TO_MANY, Connection.MANY_TO_MANY]:
                fragment_ct += ns_i_p_ct

            # Set connection.

            p._conn_type = conn_type
            p._successors = p_successors

            successors.extend(p_successors)

        for key, value in replacements.items():
            key._mask = value.mask

        # Get unique list of successors, preserving order.

        successors = Series(successors).drop_duplicates().tolist()

    # Update the record.

    record.increment_value("fragmentations", fragment_ct)
    record.increment_value("captures", capture_ct)
    record.increment_value("area_captured_sum", sum(area_captured))

    old_value = record.get_value("area_captured_max")
    old_value_is_nan = np.isnan(old_value)
    new_value_is_greater = max(area_captured) > old_value

    if old_value_is_nan or new_value_is_greater:
        record.set_value("area_captured_max", max(area_captured))

    return successors


def _create_index_map(grid, zones):
    i = np.where([z.mask for z in zones])

    index_map = np.zeros(grid.number_of_nodes, dtype=int)
    index_map[:] = -1
    index_map[i[1]] = i[0]

    return index_map


def _intersecting_zones(condition, zone_index_map, zone_list):
    mask = np.all([condition, zone_index_map > -1], 0)
    indices = np.unique(zone_index_map[np.argwhere(mask)])
    zones = [zone_list[i] for i in indices]
    return zones


def _determine_connection_type(prior_zone_count, new_zone_count):
    """Get the connection type based on the count of prior and new zones."""
    if prior_zone_count == 0:
        if new_zone_count == 1:
            return Connection.NONE_TO_ONE

    elif prior_zone_count == 1:
        if new_zone_count == 0:
            return Connection.ONE_TO_NONE
        elif new_zone_count == 1:
            return Connection.ONE_TO_ONE
        elif new_zone_count > 1:
            return Connection.ONE_TO_MANY

    elif prior_zone_count > 1:
        if new_zone_count == 1:
            return Connection.MANY_TO_ONE
        elif new_zone_count > 1:
            return Connection.MANY_TO_MANY


def _get_successors(
    p,
    conn_type,
    ps_i_ns,
    ns_i_p,
    prior_zones,
    ps_index_map,
    replacements,
    all_successors,
):
    if conn_type == Connection.ONE_TO_NONE:
        successors = []

    elif conn_type == Connection.ONE_TO_ONE:
        # The prior zone is set as the new zone because only the one new
        # and the one prior overlap.
        n = ns_i_p[0]
        replacements[p] = n
        successors = [p]

    elif conn_type in [Connection.ONE_TO_MANY, Connection.MANY_TO_MANY]:
        # Set the successors to the new zones that overlap p.
        # Although, replace the dominant n with p.

        dn = p._get_largest_intersection(ns_i_p, exclusions=list(replacements.values()))

        successors = []

        for i, n in enumerate(ns_i_p):
            dp = n._get_largest_intersection(
                ps_i_ns, exclusions=list(replacements.keys())
            )

            if n == dn and n in replacements.values():
                d = _get_replacement(replacements, n)
                successors.append(d)
            elif n == dn:
                replacements[dp] = n
                successors.append(dp)
            elif n in replacements.values():
                d = _get_replacement(replacements, n)
                successors.append(d)
            else:
                successors.append(dp)
                replacements[dp] = n

    elif conn_type == Connection.MANY_TO_ONE:
        # Set the successor to the prior zone that intersects n the most.
        n = ns_i_p[0]
        dp = n._get_largest_intersection(ps_i_ns)

        if p == dp and n in replacements.values():
            successors = [_get_replacement(replacements, n)]
        elif p == dp:
            successors = [p]
            replacements[p] = n
        elif n in replacements.values():
            successors = [_get_replacement(replacements, n)]
        else:
            successors = [dp]
            replacements[dp] = n

    return successors


def _get_replacement(replacements, new_zone):
    for key, value in replacements.items():
        if value == new_zone:
            return key


class Zone(object):
    """Zone object of SpeciesEvolver.

    The nodes and attributes of the spatial entities that taxa populate. This
    class is not intended to be managed directly.
    """

    def __init__(self, controller, mask):
        """Initialize a zone.

        Parameters
        ----------
        controller : ZoneController
            A SpeciesEvolver ZoneController.
        mask : ndarray
            The mask of the zone. True elements of this array correspond to the
            grid nodes of the zone.
        """
        self._controller = controller
        self._mask = mask.flatten()
        self._conn_type = None
        self._successors = []

    @property
    def mask(self):
        """The mask of the zone."""
        return self._mask

    @property
    def successors(self):
        """A list of zones connected to zone at the current time step."""
        return self._successors

    @property
    def area(self):
        """Zone area calculated as the sum of cell area at grid nodes."""
        area = self._controller._grid.cell_area_at_node[self._mask].sum()
        return area

    def _get_largest_intersection(self, zones, exclusions=[]):
        node_intersection_count = []
        for z in zones:
            if z in exclusions:
                node_intersection_count.append(-1)
            else:
                n = len(np.where(np.all([z.mask, self.mask], 0))[0])
                node_intersection_count.append(n)

        if all(x == -1 for x in node_intersection_count):
            return self

        return zones[np.argmax(node_intersection_count)]



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 24 16:28:00 2019

This code converts location in a link to an X, Y


@author: Jon Czuba, Katy Barnhart
"""
import numpy as np


def locate_parcel_xy(grid, parcels, parcel_time_index, parcel_number):

    # determine the location of that parcel in its link
    parcel_loc = parcels.dataset.location_in_link[
        parcel_number, parcel_time_index
    ].values

    # parcels that end their timestep off network have the starting link id
    # recorded, and np.nan as the distance.

    if not np.isnan(parcel_loc):

        # get link id
        parcel_link = int(
            parcels.dataset.element_id[parcel_number, parcel_time_index].values
        )

        # DANGER DANGER: This code assumes the verticies of links are ordered
        # from upstream to downstream. This should be the case for delineated
        # river networks, so this line should not be necessary. A quick
        # work-around is the following, but ideally verticies should be flipped in GIS.
        # parcel_loc = 0.9999 - parcel_loc

        # get the X, Y vertices of the squiggly line for that link (loaded by import_shapefile.py)

        # I am not sure these next 2 lines are necessary here, but I don't know how to do this differently and/or better.
        if "x_of_polyline" in grid.at_link:
            link_x = grid["link"]["x_of_polyline"][parcel_link]
            link_y = grid["link"]["y_of_polyline"][parcel_link]
        else:
            flow_dir = grid.at_link["flow__link_direction"][parcel_link]
            head_node = grid.node_at_link_head[parcel_link]
            tail_node = grid.node_at_link_tail[parcel_link]
            if flow_dir == -1:
                link_x = [grid.x_of_node[head_node], grid.x_of_node[tail_node]]
                link_y = [grid.y_of_node[head_node], grid.y_of_node[tail_node]]
            elif flow_dir == 1:
                # 1 = with direction of from tail to head.
                link_x = [grid.x_of_node[tail_node], grid.x_of_node[head_node]]
                link_y = [grid.y_of_node[tail_node], grid.y_of_node[head_node]]
            else:
                msg = "trying to plot on an inactive link. this should not happen."
                raise ValueError(msg)
            # eventually need to use x_of_node, y_of_node, and nodes_at_link,
            # but the upstream to downstream ordering also matters.

        # cumulative distance between squiggly-line link vertices [0 to link length]
        link_dist = np.concatenate(
            [[0], np.cumsum(np.sqrt(np.diff(link_x) ** 2 + np.diff(link_y) ** 2))]
        )

        # cumulative relative distance between squiggly-line link verticies [0 to 1]
        # divide cumulative distance by max distance to get vector of distances between
        # 0 and 1
        link_rel_dist = link_dist / np.max(link_dist)

        # # determine which two points on squiggly line bound parcel location in that link
        # upper_loc_idx = np.argmax(link_rel_dist - parcel_loc > 0)
        #
        # to_interp_link_loc = link_rel_dist[upper_loc_idx - 1 : upper_loc_idx + 1]
        # to_interp_link_x = link_x[upper_loc_idx - 1 : upper_loc_idx + 1]
        # to_interp_link_y = link_y[upper_loc_idx - 1 : upper_loc_idx + 1]
        #
        # # check values are increasing
        # np.all(np.diff(to_interp_link_loc) > 0)

        # interpolate the X,Y coordinates from the parcel location
        parcel_x = np.interp(
            parcel_loc, link_rel_dist, link_x
        )  # , left=np.nan, right=np.nan)
        parcel_y = np.interp(parcel_loc, link_rel_dist, link_y)
        # assert np.isnan(parcel_x) == False

        # save data to a single variable. better would be to save this info as an element of parcels.dataset.X and ...Y
        XY = [parcel_x, parcel_y]

        # parcels.dataset["X"] = parcel_x
        # parcels.dataset["Y"] = parcel_y

    else:
        # if that parcel is no longer in the system do not try to compute X,Y and instead return NaN
        XY = [np.nan, np.nan]

    # return the X,Y values
    return XY



# -*- coding: utf-8 -*-
"""DepthDependentTaylorNonLinearDiffuser Component.

@author: R Glade
@author: K Barnhart
@author: G Tucker
"""

import numpy as np

from landlab import Component, LinkStatus
from landlab.core.messages import deprecation_message


class DepthDependentTaylorDiffuser(Component):

    r"""
    This component implements a depth-dependent Taylor series diffusion rule,
    combining concepts of Ganti et al. (2012) and Johnstone and Hilley (2014).

    Hillslope sediment flux uses a Taylor Series expansion of the Andrews-
    Bucknam formulation of nonlinear hillslope flux derived following following
    Ganti et al., 2012 with a depth dependent component inspired Johnstone and
    Hilley (2014). The flux :math:`q_s` is given as:

    .. math::

        q_s = - K H_* \nabla \eta ( 1 + (S/S_c)^2 + (S/S_c)^4 + .. + (S/S_c)^2(n-1) ) (1 - exp( - H / H_*)

    where :math:`K` is a transport velocity coefficient, :math:`\eta` is land
    surface elevation, :math:`S` is the slope gradient (defined as
    positive downward), :math:`S_c` is the critical slope, :math:`n` is the
    number of terms, :math:`H` is the soil depth on links, and :math:`H_*` is
    the soil transport decay depth.

    The default behavior uses two terms to produce a slope dependence as
    described by Equation 6 of Ganti et al. (2012).

    This component will ignore soil thickness located at non-core nodes.

    Examples
    --------
    First lets make a simple example with flat topography.

    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import ExponentialWeatherer
    >>> from landlab.components import DepthDependentTaylorDiffuser
    >>> mg = RasterModelGrid((5, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> expweath = ExponentialWeatherer(mg)
    >>> DDdiff = DepthDependentTaylorDiffuser(mg)
    >>> expweath.calc_soil_prod_rate()
    >>> np.allclose(mg.at_node['soil_production__rate'][mg.core_nodes], 1.)
    True
    >>> DDdiff.run_one_step(2.)
    >>> np.allclose(mg.at_node['topographic__elevation'][mg.core_nodes], 0.)
    True
    >>> np.allclose(mg.at_node['bedrock__elevation'][mg.core_nodes], -2.)
    True
    >>> np.allclose(mg.at_node['soil__depth'][mg.core_nodes], 2.)
    True

    Now a more complicated example with a slope.

    >>> mg = RasterModelGrid((3, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> z += mg.node_x.copy()
    >>> BRz += mg.node_x / 2.
    >>> soilTh[:] = z - BRz
    >>> expweath = ExponentialWeatherer(mg)
    >>> DDdiff = DepthDependentTaylorDiffuser(mg)
    >>> expweath.calc_soil_prod_rate()
    >>> np.allclose(
    ...     mg.at_node['soil_production__rate'][mg.core_nodes],
    ...     np.array([ 0.60653066, 0.36787944, 0.22313016]))
    True
    >>> DDdiff.run_one_step(0.1)
    >>> np.allclose(
    ...     mg.at_node['topographic__elevation'][mg.core_nodes],
    ...     np.array([ 1.04773024, 2.02894986, 3.01755898]))
    True
    >>> np.allclose(mg.at_node['bedrock__elevation'][mg.core_nodes],
    ...     np.array([ 0.43934693, 0.96321206, 1.47768698]))
    True
    >>> np.allclose(mg.at_node['soil__depth'], z - BRz)
    True

    The DepthDependentTaylorDiffuser makes and moves soil at a rate proportional
    to slope, this means that there is a characteristic time scale for soil
    transport and an associated stability criteria for the timestep. The
    maximum characteristic time scale, :math:`De_{max}`, is given as a function of the
    hillslope diffustivity, :math:`D`, the maximum slope, :math:`S_{max}`, and the critical slope
    :math:`S_c`.

    .. math::

        De_{max} = D
            \left(
            1 +
            \left( \frac{S_{max}{S_c}\right )^2 +
            \left( \frac{S_{max}{S_c}\right )^4 +
            \dots +
            \left( \frac{S_{max}{S_c}\right )^{( 2 * ( n - 1 ))}
            \right)

    The maximum stable time step is given by

    .. math::

        dtmax = courant_factor * dx * dx / Demax

    Where the courant factor is a user defined scale (default is 0.2)

    The DepthDependentTaylorDiffuser has a boolean flag that permits a user
    to be warned if timesteps are too large for the slopes in the model grid
    (if_unstable = 'warn') and a boolean flag that turns on dynamic timestepping
    (dynamic_dt = False).

    >>> DDdiff = DepthDependentTaylorDiffuser(mg, if_unstable='warn')
    >>> DDdiff.run_one_step(2.)
    Topographic slopes are high enough such that the Courant condition is
    exceeded AND you have not selected dynamic timestepping with
    dynamic_dt=True. This may lead to infinite and/or nan values for slope,
    elevation, and soil depth. Consider using a smaller time step or dynamic
    timestepping. The Courant condition recommends a timestep of
    0.0953407607307 or smaller.

    Alternatively you can specify if_unstable='raise', and a Runtime Error will
    be raised if this condition is not met.

    Next, lets do an example with dynamic timestepping.

    >>> mg = RasterModelGrid((3, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')

    We'll use a steep slope and very little soil.

    >>> z += mg.node_x.copy()**2
    >>> BRz = z.copy() - 1.0
    >>> soilTh[:] = z - BRz
    >>> expweath = ExponentialWeatherer(mg)

    Lets try to move the soil with a large timestep. Without dynamic time
    steps, this gives a warning that we've exceeded the dynamic timestep size
    and should use a smaller timestep. We could either use the smaller timestep,
    or specify that we want to use a dynamic timestep.

    >>> DDdiff = DepthDependentTaylorDiffuser(mg, if_unstable='warn', dynamic_dt=False)
    >>> expweath.calc_soil_prod_rate()
    >>> DDdiff.run_one_step(10)
    Topographic slopes are high enough such that the Courant condition is
    exceeded AND you have not selected dynamic timestepping with
    dynamic_dt=True. This may lead to infinite and/or nan values for slope,
    elevation, and soil depth. Consider using a smaller time step or dynamic
    timestepping. The Courant condition recommends a timestep of
    0.004 or smaller.

    Now, we'll re-build the grid and do the same example with dynamic timesteps.

    >>> mg = RasterModelGrid((3, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> z += mg.node_x.copy()**2
    >>> BRz = z.copy() - 1.0
    >>> soilTh[:] = z - BRz
    >>> expweath = ExponentialWeatherer(mg)
    >>> DDdiff = DepthDependentTaylorDiffuser(mg, if_unstable='warn', dynamic_dt=True)
    >>> expweath.calc_soil_prod_rate()
    >>> DDdiff.run_one_step(10)
    >>> np.any(np.isnan(z))
    False

    Now, we'll test that changing the transport decay depth behaves as expected.

    >>> mg = RasterModelGrid((3, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> z += mg.node_x.copy()**0.5
    >>> BRz = z.copy() - 1.0
    >>> soilTh[:] = z - BRz
    >>> expweath = ExponentialWeatherer(mg)
    >>> DDdiff = DepthDependentTaylorDiffuser(mg, soil_transport_decay_depth = 0.1)
    >>> DDdiff.run_one_step(1)
    >>> soil_decay_depth_point1 = mg.at_node['topographic__elevation'][mg.core_nodes]
    >>> z[:] = 0
    >>> z += mg.node_x.copy()**0.5
    >>> BRz = z.copy() - 1.0
    >>> soilTh[:] = z - BRz
    >>> DDdiff = DepthDependentTaylorDiffuser(mg, soil_transport_decay_depth = 1.0)
    >>> DDdiff.run_one_step(1)
    >>> soil_decay_depth_1 = mg.at_node['topographic__elevation'][mg.core_nodes]
    >>> np.greater(soil_decay_depth_1[1], soil_decay_depth_point1[1])
    False

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Barnhart, K., Glade, R., Shobe, C., Tucker, G. (2019). Terrainbento 1.0: a
    Python package for multi-model analysis in long-term drainage basin
    evolution. Geoscientific Model Development  12(4), 1267--1297.
    https://dx.doi.org/10.5194/gmd-12-1267-2019

    **Additional References**

    Ganti, V., Passalacqua, P., Foufoula-Georgiou, E. (2012). A sub-grid scale
    closure for nonlinear hillslope sediment transport models Journal of
    Geophysical Research: Earth Surface  117(F2).
    https://dx.doi.org/10.1029/2011jf002181

    Johnstone, S., Hilley, G. (2015). Lithologic control on the form of
    soil-mantled hillslopes Geology  43(1), 83-86.
    https://doi.org/10.1130/G36052.1

    """

    _name = "DepthDependentTaylorDiffuser"

    _unit_agnostic = True

    _cite_as = """
    @article{barnhart2019terrain,
      author = {Barnhart, Katherine R and Glade, Rachel C and Shobe, Charles M and Tucker, Gregory E},
      title = {{Terrainbento 1.0: a Python package for multi-model analysis in long-term drainage basin evolution}},
      doi = {10.5194/gmd-12-1267-2019},
      pages = {1267---1297},
      number = {4},
      volume = {12},
      journal = {Geoscientific Model Development},
      year = {2019},
    }
    """

    _info = {
        "bedrock__elevation": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "elevation of the bedrock surface",
        },
        "soil__depth": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of soil or weathered bedrock",
        },
        "soil__flux": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m^2/yr",
            "mapping": "link",
            "doc": "flux of soil in direction of link",
        },
        "soil_production__rate": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m/yr",
            "mapping": "node",
            "doc": "rate of soil production at nodes",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__slope": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/m",
            "mapping": "link",
            "doc": "gradient of the ground surface",
        },
    }

    def __init__(
        self,
        grid,
        linear_diffusivity=None,
        slope_crit=1.0,
        soil_transport_decay_depth=1.0,
        nterms=2,
        dynamic_dt=False,
        if_unstable="pass",
        courant_factor=0.2,
        soil_transport_velocity=1.0,
    ):
        """Initialize the DepthDependentTaylorDiffuser.

        Parameters
        ----------
        grid: ModelGrid
            Landlab ModelGrid object
        linear_diffusivity: float, optional, DEPRECATED
            Hillslope diffusivity / decay depth, m/yr
            Default = 1.0
        slope_crit: float, optional
            Critical gradient parameter, m/m
            Default = 1.0
        soil_transport_decay_depth: float, optional
            characteristic transport soil depth, m
            Default = 1.0
        nterms: int, optional. default = 2
            number of terms in the Taylor expansion.
            Two terms (default) gives the behavior
            described in Ganti et al. (2012).
        dynamic_dt : bool, optional, default  = False
            Whether internal timestepping is used.
        if_unstable : str, optional, default = "pass"
            What to do if unstable (options are "pass",
            "raise", "warn")
        courant_factor : float, optional, default = 0.2
            Courant factor for timestep calculation.
        soil_transport_velocity : float, optional, default = 1.0
            Velocity parameter for soil transport, m/yr. Diffusivity is the
            product of this parameter and soil_transport_decay_depth.
        """
        super().__init__(grid)

        # Handle now-deprecated diffusivity argument
        if linear_diffusivity is None:
            self._K = soil_transport_velocity
        else:
            message = """Use of linear_diffusivity is deprecated, because the
                         name is misleading: it is actually a velocity;
                         diffusivity is obtained by multiplying by soil
                         transport decay depth. Use soil_transport_velocity
                         instead."""
            print(deprecation_message(message))
            self._K = linear_diffusivity
        self._soil_transport_decay_depth = soil_transport_decay_depth
        self._slope_crit = slope_crit
        self._nterms = nterms

        self._dynamic_dt = dynamic_dt
        self._if_unstable = if_unstable
        self._courant_factor = courant_factor

        # get reference to inputs
        self._elev = self._grid.at_node["topographic__elevation"]
        self._soil_prod_rate = self._grid.at_node["soil_production__rate"]
        self._depth = self._grid.at_node["soil__depth"]

        # create outputs if necessary and get reference.
        self.initialize_output_fields()
        self._slope = self._grid.at_link["topographic__slope"]
        self._flux = self._grid.at_link["soil__flux"]
        self._bedrock = self._grid.at_node["bedrock__elevation"]

    def soilflux(self, dt):
        """Calculate soil flux for a time period 'dt'.

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.
        """
        # establish time left as all of dt
        time_left = dt

        # begin while loop for time left
        while time_left > 0.0:

            # calculate soil__depth
            self._grid.at_node["soil__depth"][:] = (
                self._grid.at_node["topographic__elevation"]
                - self._grid.at_node["bedrock__elevation"]
            )

            # Calculate soil depth at links.
            self._H_link = self._grid.map_value_at_max_node_to_link(
                "topographic__elevation", "soil__depth"
            )

            # Calculate gradients
            self._slope = self._grid.calc_grad_at_link(self._elev)
            self._slope[self._grid.status_at_link == LinkStatus.INACTIVE] = 0.0

            # Test for time stepping courant condition
            # Test for time stepping courant condition
            courant_slope_term = 0.0
            courant_s_over_scrit = self._slope.max() / self._slope_crit
            for i in range(0, 2 * self._nterms, 2):
                courant_slope_term += courant_s_over_scrit ** i
                if np.any(np.isinf(courant_slope_term)):
                    message = (
                        "Soil flux term is infinite in Courant condition "
                        "calculation. This is likely due to "
                        "using too many terms in the Taylor expansion."
                    )
                    raise RuntimeError(message)
            # Calculate De Max
            De_max = self._K * (courant_slope_term)
            # Calculate longest stable timestep
            self._dt_max = self._courant_factor * (self._grid.dx ** 2) / De_max

            # Test for the Courant condition and print warning if user intended
            # for it to be printed.
            if (
                (self._dt_max < dt)
                and (not self._dynamic_dt)
                and (self._if_unstable != "pass")
            ):
                message = (
                    "Topographic slopes are high enough such that the "
                    "Courant condition is exceeded AND you have not "
                    "selected dynamic timestepping with dynamic_dt=True. "
                    "This may lead to infinite and/or nan values for "
                    "slope, elevation, and soil depth. Consider using a "
                    "smaller time step or dynamic timestepping. The "
                    "Courant condition recommends a timestep of "
                    "" + str(self._dt_max) + " or smaller."
                )
                if self._if_unstable == "raise":
                    raise RuntimeError(message)
                if self._if_unstable == "warn":
                    print(message)

            # if dynamic dt is selected, use it, otherwise, use the entire time
            if self._dynamic_dt:
                self._sub_dt = np.min([dt, self._dt_max])
                time_left -= self._sub_dt
            else:
                self._sub_dt = dt
                time_left = 0

            # update sed flux, topography, soil, and bedrock based on the
            # current self._sub_dt
            self._update_flux_topography_soil_and_bedrock()

    def _update_flux_topography_soil_and_bedrock(self):
        """Calculate soil flux and update topography."""
        # Calculate flux
        slope_term = 0.0
        s_over_scrit = self._slope / self._slope_crit
        for i in range(0, 2 * self._nterms, 2):
            slope_term += s_over_scrit ** i
            if np.any(np.isinf(slope_term)):
                message = (
                    "Soil flux term is infinite. This is likely due to "
                    "using too many terms in the Taylor expansion."
                )
                raise RuntimeError(message)

        self._flux[:] = -(
            (self._K * self._slope * self._soil_transport_decay_depth)
            * (slope_term)
            * (1.0 - np.exp(-self._H_link / self._soil_transport_decay_depth))
        )

        # Calculate flux divergence
        dqdx = self._grid.calc_flux_div_at_node(self._flux)

        # Calculate change in soil depth
        dhdt = self._soil_prod_rate - dqdx

        # Calculate soil depth at nodes
        self._depth[self._grid.core_nodes] += dhdt[self._grid.core_nodes] * self._sub_dt

        # prevent negative soil thickness
        self._depth[self._depth < 0.0] = 0.0

        # Calculate bedrock elevation
        self._bedrock[self._grid.core_nodes] -= (
            self._soil_prod_rate[self._grid.core_nodes] * self._sub_dt
        )

        # Update topography
        self._elev[self._grid.core_nodes] = (
            self._depth[self._grid.core_nodes] + self._bedrock[self._grid.core_nodes]
        )

    def run_one_step(self, dt):
        """

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.
        """
        self.soilflux(dt)



#! /usr/bin/env python
from enum import IntEnum, unique


@unique
class NodeStatus(IntEnum):
    """Define the boundary-type codes"""

    #: Indicate a node is *core*.
    CORE = 0
    #: Indicate a boundary node is has a fixed values.
    FIXED_VALUE = 1
    #: Indicate a boundary node is has a fixed gradient.
    FIXED_GRADIENT = 2
    #: Indicate a boundary node is wrap-around.
    LOOPED = 3
    #: Indicate a boundary node is closed
    CLOSED = 4



import numpy as np
from numpy.testing import assert_array_equal
from pytest import approx

from landlab import HexModelGrid


def test_patches_at_link():
    grid = HexModelGrid((3, 2))
    assert_array_equal(
        grid.patches_at_link,
        [
            [0, -1],
            [1, -1],
            [0, 1],
            [0, 2],
            [2, -1],
            [1, 3],
            [2, 4],
            [3, -1],
            [3, 5],
            [4, 5],
            [4, -1],
            [5, -1],
        ],
    )


def test_link_angle():
    grid = HexModelGrid((3, 2))
    assert grid.angle_of_link / np.pi * 3.0 == approx(
        [0.0, 2.0, 1.0, 2.0, 1.0, 0.0, 0.0, 1.0, 2.0, 1.0, 2.0, 0.0], rel=1.0e-5
    )



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import HexModelGrid, RasterModelGrid
from landlab.utils import get_core_node_matrix


@pytest.mark.parametrize("diff", (-1, 1))
def test_bad_value_at_node(diff):
    grid = RasterModelGrid((5, 6))
    value_at_node = np.full(grid.number_of_nodes + diff, 1.0)

    with pytest.raises(ValueError):
        get_core_node_matrix(grid, value_at_node)


@pytest.mark.parametrize("diff", (-1, 1))
def test_bad_coef_at_link(diff):
    grid = RasterModelGrid((5, 6))
    value_at_node = np.full(grid.number_of_nodes, 1.0)
    coef_at_link = np.full(grid.number_of_links + diff, 1.0)

    with pytest.raises(ValueError):
        get_core_node_matrix(grid, value_at_node, coef_at_link=coef_at_link)


def test_default_bc():
    grid = RasterModelGrid((5, 6))
    value_at_node = np.full(grid.number_of_nodes, 1.0, dtype=float)

    mat, rhs = get_core_node_matrix(grid, value_at_node)

    assert_array_equal(
        mat.toarray(),
        [
            [-4.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [1.0, -4.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, -4.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, -4.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, 0.0, -4.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 1.0, -4.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0, 1.0, -4.0, 1.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, -4.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -4.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, -4.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, -4.0, 1.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, -4.0],
        ],
    )
    assert_array_equal(
        rhs,
        [
            [-2.0],
            [-1.0],
            [-1.0],
            [-2.0],
            [-1.0],
            [0.0],
            [0.0],
            [-1.0],
            [-2.0],
            [-1.0],
            [-1.0],
            [-2.0],
        ],
    )


def test_small_grids():
    grid = RasterModelGrid((3, 3))
    value_at_node = np.full(grid.number_of_nodes, 1.0)
    mat, rhs = get_core_node_matrix(grid, value_at_node)
    assert_array_equal(mat.toarray(), [[-4.0]])
    assert_array_equal(rhs, [[-4.0]])

    grid = RasterModelGrid((4, 4))
    value_at_node = np.full(grid.number_of_nodes, 1.0)
    mat, rhs = get_core_node_matrix(grid, value_at_node)

    assert_array_equal(
        mat.toarray(),
        [
            [-4.0, 1.0, 1.0, 0.0],
            [1.0, -4.0, 0.0, 1.0],
            [1.0, 0.0, -4.0, 1.0],
            [0.0, 1.0, 1.0, -4.0],
        ],
    )
    assert_array_equal(rhs, [[-2.0], [-2.0], [-2.0], [-2.0]])


def test_with_fixed_value_bc():
    grid = RasterModelGrid((4, 5))
    grid.status_at_node[13] = grid.BC_NODE_IS_FIXED_VALUE
    grid.status_at_node[2] = grid.BC_NODE_IS_CLOSED

    value_at_node = np.arange(grid.number_of_nodes, dtype=float)

    mat, rhs = get_core_node_matrix(grid, value_at_node)
    assert_array_equal(
        mat.toarray(),
        [
            [-4.0, 1.0, 0.0, 1.0, 0.0],
            [1.0, -3.0, 1.0, 0.0, 1.0],
            [0.0, 1.0, -4.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, -4.0, 1.0],
            [0.0, 1.0, 0.0, 1.0, -4.0],
        ],
    )
    assert_array_equal(
        rhs,
        [[-6.0], [0.0], [-25.0], [-26.0], [-30.0]],
    )


def test_with_coef():
    grid = RasterModelGrid((4, 5))
    grid.status_at_node[13] = grid.BC_NODE_IS_FIXED_VALUE
    grid.status_at_node[2] = grid.BC_NODE_IS_CLOSED

    value_at_node = np.arange(grid.number_of_nodes, dtype=float)

    coef_at_link = np.arange(grid.number_of_links, dtype=np.double)
    mat, rhs = get_core_node_matrix(grid, value_at_node, coef_at_link=coef_at_link)
    assert_array_equal(
        mat.toarray(),
        [
            [-38.0, 10.0, 0.0, 14.0, 0.0],
            [10.0, -36.0, 11.0, 0.0, 15.0],
            [0.0, 11.0, -46.0, 0.0, 0.0],
            [14.0, 0.0, 0.0, -74.0, 19.0],
            [0.0, 15.0, 0.0, 19.0, -78.0],
        ],
    )
    assert_array_equal(rhs, [[-6.0], [0.0], [-25.0], [-26.0], [-30.0]])


@pytest.mark.parametrize("Grid", (RasterModelGrid, HexModelGrid))
def test_scalar_value_at_node(Grid):
    grid = Grid((4, 5))

    _, expected = get_core_node_matrix(grid, np.full(grid.number_of_nodes, 11.0))
    _, actual = get_core_node_matrix(grid, 11.0)

    assert_array_equal(expected, actual)


@pytest.mark.parametrize("Grid", (RasterModelGrid, HexModelGrid))
def test_scalar_coef_at_link(Grid):
    grid = Grid((4, 5))

    expected, _ = get_core_node_matrix(
        grid, 1.0, coef_at_link=np.full(grid.number_of_links, 11.0)
    )
    actual, _ = get_core_node_matrix(grid, 1.0, coef_at_link=11.0)

    assert_array_equal(expected.toarray(), actual.toarray())


@pytest.mark.parametrize("Grid", (RasterModelGrid, HexModelGrid))
def test_default_coef_is_one(Grid):
    grid = Grid((4, 5))
    grid.status_at_node[13] = grid.BC_NODE_IS_FIXED_VALUE
    grid.status_at_node[2] = grid.BC_NODE_IS_CLOSED

    value_at_node = np.arange(grid.number_of_nodes, dtype=float)

    coef_at_link = np.full(grid.number_of_links, 1.0, dtype=float)

    expected, _ = get_core_node_matrix(grid, value_at_node, coef_at_link=coef_at_link)
    actual, _ = get_core_node_matrix(grid, value_at_node)

    assert_array_equal(expected.toarray(), actual.toarray())


@pytest.mark.parametrize("Grid", (RasterModelGrid, HexModelGrid))
def test_value_at_node(Grid):
    grid = Grid((4, 5))

    value_at_node = np.full(grid.number_of_nodes, 1.0)
    coef_at_link = np.full(grid.number_of_links, 1.0)

    mat, rhs = get_core_node_matrix(grid, value_at_node, coef_at_link=coef_at_link)
    actual, actual_rhs = get_core_node_matrix(grid, value_at_node * 2, coef_at_link)

    assert_array_equal(mat.toarray(), actual.toarray())
    assert_array_equal(rhs, actual_rhs / 2.0)


@pytest.mark.parametrize("Grid", (RasterModelGrid, HexModelGrid))
def test_coef_at_link(Grid):
    grid = Grid((4, 5))

    value_at_node = np.full(grid.number_of_nodes, 1.0)
    coef_at_link = np.full(grid.number_of_links, 1.0)

    mat, rhs = get_core_node_matrix(grid, value_at_node, coef_at_link=coef_at_link)
    actual, actual_rhs = get_core_node_matrix(grid, value_at_node, coef_at_link * 2.0)

    assert_array_equal(mat.toarray(), actual.toarray() / 2.0)
    assert_array_equal(rhs, actual_rhs)


def test_hex_grid():
    grid = HexModelGrid((4, 3))
    value_at_node = np.full(grid.number_of_nodes, 1.0)
    coef_at_link = np.full(grid.number_of_links, 1.0)

    mat, rhs = get_core_node_matrix(grid, value_at_node, coef_at_link=coef_at_link)

    assert_array_equal(
        mat.toarray(),
        [
            [-6.0, 1.0, 1.0, 1.0, 0.0],
            [1.0, -6.0, 0.0, 1.0, 1.0],
            [1.0, 0.0, -6.0, 1.0, 0.0],
            [1.0, 1.0, 1.0, -6.0, 1.0],
            [0.0, 1.0, 0.0, 1.0, -6.0],
        ],
    )
    assert_array_equal(rhs, [[-3.0], [-3.0], [-4.0], [-2.0], [-4.0]])



#!/usr/bin/env python3

"""unit tests for landlab.io.obj module"""
import pathlib

import pytest

from landlab import HexModelGrid, RasterModelGrid
from landlab.io import write_obj

LITTLE_HEX_OBJ = """# landlabgrid
#
g landlabgrid
v 1.0 0.0 0.0
v 3.0 0.0 0.0
v 0.0 1.732051 0.0
v 2.0 1.732051 1.0
v 4.0 1.732051 0.0
v 1.0 3.464102 0.0
v 3.0 3.464102 0.0
f 4// 1// 2//
f 4// 3// 1//
f 5// 4// 2//
f 6// 3// 4//
f 7// 4// 5//
f 7// 6// 4//
"""

LITTLE_RAST_OBJ = """# landlabgrid
#
g landlabgrid
v 0.0 0.0 0.0
v 2.0 0.0 0.0
v 4.0 0.0 0.0
v 0.0 2.0 0.0
v 2.0 2.0 1.0
v 4.0 2.0 0.0
v 0.0 4.0 0.0
v 2.0 4.0 0.0
v 4.0 4.0 0.0
f 5// 4// 1//
f 1// 2// 5//
f 6// 5// 2//
f 2// 3// 6//
f 8// 7// 4//
f 4// 5// 8//
f 9// 8// 5//
f 5// 6// 9//
"""


def test_write_to_filelike(tmpdir):
    grid = RasterModelGrid((3, 3), xy_spacing=2.0)
    z = grid.add_zeros("topographic__elevation", at="node")
    z[4] = 1.0

    with tmpdir.as_cwd():
        with open("test_quad.obj", "w") as fp:
            write_obj(fp, grid)
        with open("test_quad.obj", "r") as fp:
            assert fp.read() == LITTLE_RAST_OBJ


@pytest.mark.parametrize("fname", (pathlib.Path("test_hex.obj"), "test_hex.obj"))
def test_write_hex_to_path(tmpdir, fname):
    grid = HexModelGrid((3, 2), spacing=2.0)
    z = grid.add_zeros("topographic__elevation", at="node")
    z[3] = 1.0

    with tmpdir.as_cwd():
        actual = write_obj(fname, grid)
        assert actual == fname
        with open(fname, "r") as fp:
            assert fp.read() == LITTLE_HEX_OBJ


def test_write_raster(tmpdir):
    grid = RasterModelGrid((3, 3), xy_spacing=2.0)
    z = grid.add_zeros("topographic__elevation", at="node")
    z[4] = 1.0

    with tmpdir.as_cwd():
        write_obj("test_quad.obj", grid)
        with open("test_quad.obj", "r") as fp:
            assert fp.read() == LITTLE_RAST_OBJ


def test_field_name(tmpdir):
    grid = RasterModelGrid((3, 3), xy_spacing=2.0)
    z = grid.add_zeros("z", at="node")
    z[4] = 1.0

    with tmpdir.as_cwd():
        write_obj("test_quad.obj", grid, field_for_z="z")
        with open("test_quad.obj", "r") as fp:
            assert fp.read() == LITTLE_RAST_OBJ


def test_clobber(tmpdir):
    grid = RasterModelGrid((3, 3), xy_spacing=2.0)
    z = grid.add_zeros("topographic__elevation", at="node")
    z[4] = 1.0

    with tmpdir.as_cwd():
        with open("test_quad.obj", "w") as fp:
            pass

        with pytest.raises(ValueError):
            write_obj("test_quad.obj", grid)

        write_obj("test_quad.obj", grid, clobber=True)
        with open("test_quad.obj", "r") as fp:
            assert fp.read() == LITTLE_RAST_OBJ



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 08 13:50:43 2021

@author: BenjaminCampforts


Doc tests and unit tests for bedrock landslides using the BedrockLandslider component.
"""

import numpy as np
import pytest
from numpy import testing

from landlab import FieldError, RasterModelGrid
from landlab.components import BedrockLandslider, PriorityFloodFlowRouter


def test_inputFields_flowRouter():
    """
    BedrockLandslider should throw an error when topograhy is not equal to the sum of
    bedrock and soil thickness
    """
    # Make a raster model grid and create a plateau
    n_rows = 5
    n_columns = 5
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    b = mg.add_zeros("bedrock__elevation", at="node")

    # make plateau at 10m
    b += 10

    PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)

    # instantiate the slider
    with pytest.raises(RuntimeError):
        BedrockLandslider(mg)


def test_inputFields_soil():
    """
    BedrockLandslider should throw an error when the soil__depth field is not provided
    """
    n_rows = 5
    n_columns = 5
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    mg.add_zeros("topographic__elevation", at="node")
    # mg.add_zeros("soil__depth", at='node')
    mg.add_zeros("bedrock__elevation", at="node")

    PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)

    # instantiate the slider
    with pytest.raises(FieldError):
        BedrockLandslider(mg)


def test_inputFields_bedrock():
    """
    BedrockLandslider should create the bedrock__elevation field when not provided
    """
    n_rows = 5
    n_columns = 5
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    # mg.add_zeros("bedrock__elevation", at='node')

    PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)
    BedrockLandslider(mg)

    # instantiate the slider
    assert "bedrock__elevation" in mg.at_node


def test_properties_phi_fraction_fines_LS():
    """
    BedrockLandslider should throw an error when phi/fraction_fines_LS < 0 or phi > 0
    """
    # Make a raster model grid and create a plateau
    n_rows = 5
    n_columns = 5
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    mg.add_zeros("topographic__elevation", at="node")
    mg.add_zeros("soil__depth", at="node")
    mg.add_zeros("bedrock__elevation", at="node")

    PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)

    # instantiate the slider
    with pytest.raises(ValueError):
        BedrockLandslider(mg, phi=-0.2)
    # instantiate the slider
    with pytest.raises(ValueError):
        BedrockLandslider(mg, phi=1.2)
    # instantiate the slider
    with pytest.raises(ValueError):
        BedrockLandslider(mg, fraction_fines_LS=-0.2)
    # instantiate the slider
    with pytest.raises(ValueError):
        BedrockLandslider(mg, fraction_fines_LS=1.2)


def test_sliding_plain():
    """
    Test that if BedrockLandslider maths are sound and follow the Culmann theory
    """
    # Make a raster model grid and create a plateau
    n_rows = 5
    n_columns = 5
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    z = mg.add_zeros("topographic__elevation", at="node")
    s = mg.add_zeros("soil__depth", at="node")
    b = mg.add_zeros("bedrock__elevation", at="node")

    # make plateau at 10m
    b += 10
    # Lower boundary cell to 0
    b[2] = 0
    z[:] = b + s
    fd = PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)
    hy = BedrockLandslider(mg, landslides_return_time=1)

    # run flow director and BedrockLandslider for one timestep
    fd.run_one_step()
    hy.run_one_step(dt=1)
    # After one timestep, we can predict eactly where the landslide will occur.
    # The return time is set to 1 year so that probability for sliding is 100%
    # The angle of internal driction is 1m/m, the topographical gradient is 10 m/m
    # At cardinal cells, the sliding plain will be at (1+10)/2 = 5.5 m/m.
    # With a dx of 1, the cardial cell next to the critical sliding node must
    # be 5.5 m and the diagonal one at 5.5 * sqrt(2) = 7.8 m
    err_msg = "Error in the calculation of the sliding plain"
    testing.assert_almost_equal(
        [5.5 * np.sqrt(2), 5.5, 5.5 * np.sqrt(2)], z[6:9], decimal=5, err_msg=err_msg
    )


def test_sliding_evolution():
    """
    Test that if BedrockLandslider is run for a long enough time, slopes evolve to the
    angle of internal friction
    """
    # Make a raster model grid and create a plateau
    n_rows = 5
    n_columns = 5
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    z = mg.add_zeros("topographic__elevation", at="node")
    s = mg.add_zeros("soil__depth", at="node")
    b = mg.add_zeros("bedrock__elevation", at="node")

    # make plateau at 10m
    b += 10
    # Lower boundary cell to 0
    b[2] = 0
    z[:] = b + s
    fd = PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)

    # Instanciate the slider, but this time provide the node at which landslides
    # should occur (node 2)
    hy = BedrockLandslider(
        mg,
        angle_int_frict=1,
        landslides_return_time=0.01,
        landslides_on_boundary_nodes=True,
        critical_sliding_nodes=[2],
    )

    # run flow director and BedrockLandslider for long enough so that surface evolves
    # towards the angle of internal friction timestep

    for _ in range(50):
        fd.run_one_step()
        hy.run_one_step(dt=1)

    # After 50 iterations, landslide erosion starting fromnode 2 evolves to a
    # known slope, equal to the angle of internal friction, multiplied by the
    # distance from node 2 giving:
    topo_cal = np.array(
        [
            [2.0, 1.0, 0.0, 1.0, 2.0],
            [2.23606798, 1.41421356, 1.0, 1.41421356, 2.23606798],
            [2.82842712, 2.23606798, 2.0, 2.23606798, 2.82842712],
            [3.60555128, 3.16227766, 3.0, 3.16227766, 3.60555128],
            [4.47213595, 4.12310563, 4.0, 4.12310563, 4.47213595],
        ]
    ).flatten()

    err_msg = "Slope is not evolving to theoretical value"
    testing.assert_almost_equal(topo_cal, z, decimal=5, err_msg=err_msg)


def test_boundary_nodes():
    # %%
    """
    Test that if BedrockLandslider cannot make or initate landslides at boundary nodes,
    it doesn't
    """
    # Make a raster model grid and create a plateau
    n_rows = 5
    n_columns = 5
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    z = mg.add_zeros("topographic__elevation", at="node")
    s = mg.add_zeros("soil__depth", at="node")
    b = mg.add_zeros("bedrock__elevation", at="node")

    # make plateau at 10m
    b += 10
    # Lower boundary cell to 0
    b[2] = 0
    z[:] = b + s
    ini_topo_bd = z[mg.boundary_nodes]
    fd = PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)

    # Instanciate the slider, but this time provide the node at which landslides
    # should occur (node 2)
    hy = BedrockLandslider(
        mg,
        angle_int_frict=1,
        landslides_return_time=0.01,
        landslides_on_boundary_nodes=False,
        critical_sliding_nodes=[2],
    )

    # run flow director and BedrockLandslider for long enough so that surface evolves
    # towards the angle of internal friction timestep

    for _ in range(5):
        fd.run_one_step()
        hy.run_one_step(dt=1)
    # Final topo at boundary nodes should be inital topo
    err_msg = "No landslide erosion allowed at boundary nodes"

    testing.assert_almost_equal(
        ini_topo_bd, z[mg.boundary_nodes], decimal=5, err_msg=err_msg
    )


# %%
def test_mass_balance_noporosity():
    """
    Test is mass balance is conserved during sliding events.
    First test for soils with zero porosity and where none of the sediment is
    evacuated as suspended sediment.
    """
    # Make a raster model grid and create a plateau
    n_rows = 9
    n_columns = 9
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    z = mg.add_zeros("topographic__elevation", at="node")
    s = mg.add_zeros("soil__depth", at="node")
    b = mg.add_zeros("bedrock__elevation", at="node")

    # make plateau at 10m
    b += 10
    # Elevate central node
    b[40] = 10
    # Create slope at margin so that sediment will be evacuated
    b[mg.boundary_nodes] = 0
    z[:] = b + s
    fd = PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)

    # Instanciate the slider, but this time provide the node at which landslides
    # should occur (node 2)
    hy = BedrockLandslider(
        mg,
        angle_int_frict=1,
        landslides_return_time=1,
        landslides_on_boundary_nodes=True,
        phi=0,
        fraction_fines_LS=0,
        verbose_landslides=True,
    )

    # Assume equal bulk density for rock and soil
    total_vol_before = np.sum(b) + np.sum(s)

    # Keep track of total volumes evacuated
    vol_SSY_tot = 0
    V_leaving_tot = 0
    # Run for some time
    for _ in range(5):
        fd.run_one_step()
        vol_SSY, V_leaving = hy.run_one_step(dt=1)
        vol_SSY_tot += vol_SSY
        V_leaving_tot += V_leaving

    # Check if list with landslide properties is generated
    assert isinstance(hy.landslides_size, list)
    assert isinstance(hy.landslides_volume, list)
    assert isinstance(hy.landslides_volume_bed, list)
    assert isinstance(hy.landslides_volume_sed, list)

    total_vol_after = np.sum(b) + np.sum(s) + vol_SSY_tot + V_leaving_tot
    # Check mass balance
    err_msg = "Mass balance not okey"
    testing.assert_almost_equal(
        total_vol_before, total_vol_after, decimal=5, err_msg=err_msg
    )


def test_mass_balance_porosity():
    """
    Test is mass balance is conserved during sliding events.
    Test for soils with given porosity phi and where none of the sediment is
    evacuated as suspended sediment.
    """
    # Make a raster model grid and create a plateau
    n_rows = 9
    n_columns = 9
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    z = mg.add_zeros("topographic__elevation", at="node")
    s = mg.add_zeros("soil__depth", at="node")
    b = mg.add_zeros("bedrock__elevation", at="node")

    # make plateau at 10m
    b += 10
    # Elevate central node
    b[40] = 10
    # Create slope at margin so that sediment will be evacuated
    b[mg.boundary_nodes] = 0
    z[:] = b + s
    fd = PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)

    # Instanciate the slider, but this time provide the node at which landslides
    # should occur (node 2)
    phi = 0.3
    hy = BedrockLandslider(
        mg,
        angle_int_frict=1,
        landslides_return_time=1,
        landslides_on_boundary_nodes=True,
        phi=phi,
        fraction_fines_LS=0,
    )

    # Assume equal bulk density for rock and soil
    total_vol_before = np.sum(b) + np.sum(s) * (1 - phi)

    # Keep track of total volumes evacuated
    vol_SSY_tot = 0
    V_leaving_tot = 0
    # Run for some time
    for _ in range(5):
        fd.run_one_step()
        vol_SSY, V_leaving = hy.run_one_step(dt=1)
        vol_SSY_tot += vol_SSY
        V_leaving_tot += V_leaving

    total_vol_after = np.sum(b) + np.sum(s) * (1 - phi) + vol_SSY_tot + V_leaving_tot
    # Check mass balance
    err_msg = "Mass balance not okey"
    testing.assert_almost_equal(
        total_vol_before, total_vol_after, decimal=5, err_msg=err_msg
    )


def test_mass_balance_porosity_suspension():
    """
    Test is mass balance is conserved during sliding events.
    Test for soils with given porosity phi and where none of the sediment is
    evacuated as suspended sediment.
    """
    # %%
    # Make a raster model grid and create a plateau
    n_rows = 9
    n_columns = 9
    spacing = 1
    mg = RasterModelGrid((n_rows, n_columns), xy_spacing=spacing)
    z = mg.add_zeros("topographic__elevation", at="node")
    s = mg.add_zeros("soil__depth", at="node")
    b = mg.add_zeros("bedrock__elevation", at="node")

    # make plateau at 10m
    b += 10
    # Elevate central node
    b[40] = 10
    # Create slope at margin so that sediment will be evacuated
    b[mg.boundary_nodes] = 0
    z[:] = b + s
    fd = PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)

    # Instanciate the slider, but this time provide the node at which landslides
    # should occur (node 2)
    phi = 0.3
    fraction_fines_LS = 0.5
    hy = BedrockLandslider(
        mg,
        angle_int_frict=1,
        landslides_return_time=1,
        landslides_on_boundary_nodes=True,
        phi=phi,
        fraction_fines_LS=fraction_fines_LS,
    )

    # Assume equal bulk density for rock and soil
    total_vol_before = np.sum(b) + np.sum(s) * (1 - phi)

    # Keep track of total volumes evacuated
    vol_SSY_tot = 0
    V_leaving_tot = 0
    # Run for some time
    for _ in range(5):
        fd.run_one_step()
        vol_SSY, V_leaving = hy.run_one_step(dt=1)
        vol_SSY_tot += vol_SSY
        V_leaving_tot += V_leaving

    total_vol_after = np.sum(b) + np.sum(s) * (1 - phi) + vol_SSY_tot + V_leaving_tot
    # Check mass balance
    err_msg = "Mass balance not okey"
    testing.assert_almost_equal(
        total_vol_before, total_vol_after, decimal=5, err_msg=err_msg
    )



import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free


DTYPE = int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_perimeter_nodes_rect_horizontal(
    shape,
    np.ndarray[DTYPE_t, ndim=1] perimeter_nodes,
):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_nodes = n_rows * n_cols
    cdef int i
    cdef int node

    # Right edge
    i = 0
    for node in range(n_cols - 1, n_nodes - 1, n_cols):
        perimeter_nodes[i] = node
        i += 1

    # Top edge
    for node in range(n_nodes - 1, n_nodes - n_cols, - 1):
        perimeter_nodes[i] = node
        i += 1

    # Left edge
    for node in range((n_rows - 1) * n_cols, 0, - n_cols):
        perimeter_nodes[i] = node
        i += 1

    # Bottom edge
    for node in range(0, n_cols - 1):
        perimeter_nodes[i] = node
        i += 1


@cython.boundscheck(False)
def fill_perimeter_nodes_rect_vertical(
    shape,
    np.ndarray[DTYPE_t, ndim=1] perimeter_nodes,
):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_nodes = n_rows * n_cols
    cdef int i
    cdef int node
    cdef int offset_to_right_edge
    cdef int offset_to_top_edge
    cdef int offset_to_left_edge
    cdef int offset_to_bottom_edge

    offset_to_top_edge = n_rows - 1
    offset_to_left_edge = offset_to_top_edge + n_cols - 1
    offset_to_bottom_edge = offset_to_left_edge + n_rows - 1
    offset_to_right_edge = offset_to_bottom_edge + n_cols - 1

    if n_cols % 2 == 0:
        perimeter_nodes[0] = n_cols - 1
    else:
        perimeter_nodes[0] = n_cols // 2

    # Right edge
    for i in range(1, offset_to_top_edge + 1):
        perimeter_nodes[i] = perimeter_nodes[i - 1] + n_cols

    # Top edge
    perimeter_nodes[offset_to_top_edge + 1] = perimeter_nodes[offset_to_top_edge]
    if n_cols % 2 == 0:
        perimeter_nodes[offset_to_top_edge + 1] -= n_cols // 2
    else:
        perimeter_nodes[offset_to_top_edge + 1] += n_cols // 2
    for i in range(offset_to_top_edge + 2, offset_to_left_edge + 1, 2):
        perimeter_nodes[i] = perimeter_nodes[i - 2] - 1
        perimeter_nodes[i + 1] = perimeter_nodes[i - 1] - 1

    # Left edge
    for i in range(offset_to_left_edge + 1, offset_to_bottom_edge + 1):
        perimeter_nodes[i] = perimeter_nodes[i - 1] - n_cols

    # Bottom edge
    perimeter_nodes[offset_to_bottom_edge] = 0
    for i in range(offset_to_bottom_edge + 2, offset_to_right_edge, 2):
        perimeter_nodes[i] = perimeter_nodes[i - 2] + 1

    perimeter_nodes[offset_to_bottom_edge + 1] = (n_cols + 1) // 2
    for i in range(offset_to_bottom_edge + 3, offset_to_right_edge, 2):
        perimeter_nodes[i] = perimeter_nodes[i - 2] + 1


@cython.boundscheck(False)
def fill_perimeter_nodes_hex_horizontal(
    shape,
    np.ndarray[DTYPE_t, ndim=1] perimeter_nodes,
):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int longest_row = n_rows // 2
    cdef int offset_to_right_edge
    cdef int offset_to_top_edge
    cdef int offset_to_left_edge
    cdef int offset_to_bottom_edge
    cdef int i, i0
    cdef int node
    cdef int row
    cdef int * nodes_per_row

    try:
        nodes_per_row = <int *>malloc(n_rows * sizeof(int))

        nodes_per_row[0] = n_cols
        for row in range(1, longest_row + 1):
            nodes_per_row[row] = nodes_per_row[row - 1] + 1
        for row in range(longest_row + 1, n_rows):
            nodes_per_row[row] = nodes_per_row[row - 1] - 1

        offset_to_top_edge = n_rows - 1
        offset_to_left_edge = offset_to_top_edge + nodes_per_row[n_rows - 1] - 1
        offset_to_bottom_edge = offset_to_left_edge + n_rows - 1
        offset_to_right_edge = offset_to_bottom_edge + n_cols - 1

        perimeter_nodes[0] = n_cols - 1

        # Right edge
        row = 1
        for i in range(1, offset_to_top_edge + 1):
            perimeter_nodes[i] = perimeter_nodes[i - 1] + nodes_per_row[row]
            row += 1

        # Top edge
        for i in range(offset_to_top_edge + 1, offset_to_left_edge + 1):
            perimeter_nodes[i] = perimeter_nodes[i - 1] - 1

        # Left edge
        row = n_rows - 2
        for i in range(offset_to_left_edge + 1, offset_to_bottom_edge + 1):
            perimeter_nodes[i] = perimeter_nodes[i - 1] - nodes_per_row[row]
            row -= 1

        # Bottom edge
        for i in range(offset_to_bottom_edge + 1, offset_to_right_edge):
            perimeter_nodes[i] = perimeter_nodes[i - 1] + 1
    finally:
        free(nodes_per_row)


@cython.boundscheck(False)
def fill_perimeter_nodes_hex_vertical(
    shape,
    np.ndarray[DTYPE_t, ndim=1] perimeter_nodes,
):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int longest_col = n_cols // 2
    cdef int max_nodes_per_row = n_cols - n_cols // 2
    cdef int offset_to_right_edge
    cdef int offset_to_top_edge
    cdef int offset_to_left_edge
    cdef int offset_to_bottom_edge
    cdef int i, i0
    cdef int node
    cdef int row

    if n_cols % 2 == 1:
        offset_to_top_edge = n_rows - 1
    else:
        offset_to_top_edge = n_rows
    offset_to_left_edge = offset_to_top_edge + n_cols - 1
    offset_to_bottom_edge = offset_to_left_edge + n_rows - 1
    offset_to_right_edge = offset_to_bottom_edge + n_cols - 1

    # Bottom edge
    nodes_per_row = 1
    perimeter_nodes[offset_to_bottom_edge + longest_col] = 0
    for i in range(offset_to_bottom_edge + longest_col + 1, offset_to_right_edge):
        nodes_per_row += 1
        perimeter_nodes[i] = perimeter_nodes[i - 1] + nodes_per_row

    perimeter_nodes[0] = perimeter_nodes[offset_to_right_edge - 1] + max_nodes_per_row

    # Right edge
    for i in range(1, offset_to_top_edge + 1):
        perimeter_nodes[i] = perimeter_nodes[i - 1] + n_cols

    # Top edge
    nodes_per_row = max_nodes_per_row
    for i in range(offset_to_top_edge + 1, offset_to_left_edge - longest_col + 1):
        nodes_per_row -= 1
        perimeter_nodes[i] = perimeter_nodes[i - 1] + nodes_per_row
    nodes_per_row = 1
    for i in range(offset_to_left_edge - longest_col + 1, offset_to_left_edge):
        nodes_per_row += 1
        perimeter_nodes[i] = perimeter_nodes[i - 1] - nodes_per_row
    perimeter_nodes[offset_to_left_edge] = (
        perimeter_nodes[offset_to_left_edge - 1] - max_nodes_per_row
    )

    # Left edge
    for i in range(offset_to_left_edge + 1, offset_to_bottom_edge + 1):
        perimeter_nodes[i] = perimeter_nodes[i - 1] - n_cols

    # Bottom edge
    perimeter_nodes[offset_to_bottom_edge + 1] = (
        perimeter_nodes[offset_to_bottom_edge] - nodes_per_row
    )
    nodes_per_row = max_nodes_per_row
    for i in range(offset_to_bottom_edge + 2, offset_to_bottom_edge + longest_col):
        nodes_per_row -= 1
        perimeter_nodes[i] = perimeter_nodes[i - 1] - nodes_per_row



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.data_record.data_record.DataRecord
Dimension = item_id

Last updated 8/24/2018

"""

import numpy as np
import pytest

from landlab import RasterModelGrid

grid = RasterModelGrid((3, 3))
shape = (3, 3)
my_items2 = {
    "grid_element": np.array(("node", "link"), dtype=str),
    "element_id": np.array([1, 3]),
}


def test_dr_item_name(dr_item):
    assert dr_item._name == "DataRecord"


def test_grid_shape(dr_item):
    assert dr_item._grid.number_of_node_rows == shape[0]
    assert dr_item._grid.number_of_node_columns == shape[1]


def test_permitted_locations(dr_item):
    assert dr_item._permitted_locations == grid.groups


def test_coordinates(dr_item):
    assert len(dr_item.dataset.dims) == 1
    assert list(dr_item.dataset.item_id.values) == [0, 1]
    assert list(dr_item.item_coordinates) == [0, 1]
    assert dr_item.number_of_items == len(my_items2["element_id"])
    with pytest.raises(AttributeError):
        dr_item.dataset.time
    with pytest.raises(AttributeError):
        dr_item.time_coordinates
    with pytest.raises(AttributeError):
        dr_item.number_of_timesteps
    with pytest.raises(AttributeError):
        dr_item.earliest_time
    with pytest.raises(AttributeError):
        dr_item.latest_time
    with pytest.raises(AttributeError):
        dr_item.prior_time


def test_variable_names(dr_item):
    assert sorted(dr_item.variable_names) == sorted(["grid_element", "element_id"])


def test_add_item(dr_item):
    dr_item.add_item(
        new_item={
            "grid_element": np.array(["node", "node"]),
            "element_id": np.array([4, 4]),
        },
        new_item_spec={"size": (["item_id"], [10, 5])},
    )
    assert (
        dr_item.dataset["grid_element"].values[3],
        dr_item.dataset["element_id"].values[3],
        dr_item.dataset["size"].values[3],
    ) == ("node", 4.0, 5.0)


def test_get_data(dr_item):
    assert dr_item.get_data(item_id=[1], data_variable="grid_element") == "link"
    assert dr_item.get_data(data_variable="element_id")[1] == 3


def test_set_data(dr_item):
    dr_item.set_data(item_id=[1], data_variable="element_id", new_value=2)
    assert dr_item.dataset["element_id"].values[1] == 2
